from .UIFormWidget import UIFormFactory # isort:skip (prereq for FormDialog)
from .FormDialog import FormDialog
from .ProgressTimerDialog import ProgressTimerDialog
from .UIMultiStepWidget import UIMultiStepFactory

__all__ = ['FormDialog', 'ProgressTimerDialog', 'UIFormFactory', 'UIMultiStepFactory']



from PySide2 import QtWidgets
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QHBoxLayout, QListWidget, QStackedWidget, QVBoxLayout, QWidget

from .UIFormWidget import UIFormFactory


class UIStackedWidget(object):
    '''
    By default:

             QWidget or QDockWidget
    +----------------------------------------------------------+
    |        QVBoxLayout                                       |
    |   +---------------------------------------------------+  |
    |   |                                                   |  |
    |   |    +-------------------------------------------+  |  |
    |   |    |   QList                                   |  |  |
    |   |    |                                           |  |  |
    |   |    |                                           |  |  |
    |   |    +-------------------------------------------+  |  |
    |   |    +-------------------------------------------+  |  |
    |   |    |   QStackedWidget                          |  |  |
    |   |    |                                           |  |  |
    |   |    |                                           |  |  |
    |   |    +-------------------------------------------+  |  |
    |   |                                                   |  |
    |   +---------------------------------------------------+  |
    |                                                          |
    +----------------------------------------------------------+

    or if layout = horizontal:

             QWidget or QDockWidget
    +----------------------------------------------------------+
    |        QHBoxLayout                                       |
    |   +---------------------------------------------------+  |
    |   |                                                   |  |
    |   |    +--------------+   +------------------------+  |  |
    |   |    |   QList      |   |  QStackedWidget        |  |  |
    |   |    |              |   |                        |  |  |
    |   |    |              |   |                        |  |  |
    |   |    +--------------+   +------------------------+  |  |
    |   |                                                   |  |
    |   +---------------------------------------------------+  |
    |                                                          |
    +----------------------------------------------------------+


    '''
    def createStack(self, layout='vertical'):
        self.stack_list = QListWidget()
        self.list_margin_size = 11

        self.Stack = QStackedWidget(self)
        self.layout_type = layout

        if self.layout_type == 'horizontal':
            box = QHBoxLayout(self)
            self.list_margin_size = 18
            self.stack_list.setStyleSheet(
                "margin-top: {size}px".format(size=self.list_margin_size))
        else:
            box = QVBoxLayout(self)
            self.stack_list.setStyleSheet(
                "margin-left : {size}px; margin-right : {size2}px".format(
                    size=self.list_margin_size, size2=self.list_margin_size + 1))

        box.addWidget(self.stack_list)
        box.addWidget(self.Stack)
        box.setAlignment(self.stack_list, Qt.AlignTop)

        self.setLayout(box)
        self.stack_list.currentRowChanged.connect(self.display)

        self.tabs = {}

        self.num_tabs = 0
        self.widgets = {}

    def display(self, i):
        self.Stack.setCurrentIndex(i)

    def currentIndex(self):
        return self.Stack.currentIndex()

    def addTab(self, label, widget='form', title=None, number_title=True):
        ''' Adds a tab to the StackedWidget.

        Parameters
        ----------
        label: str
            The name that will be used to refer to the tab.
        title: str
            The name the tab will be listed as in the QListWidget.
        number_title: bool, default True
            Determines whether the title of the tab is generated
            by adding a number in front of the label. E.g. if the
            label is "FBP" and this is the first tab that has been
            added then the title will be "1 - FBP" if number_title
            has been set to True and title=None
        widget: QWidget or str: 'form', default 'form'
            the widget that will be contained in the tab.
            by default this is set to 'form' which means an empty
            FormWidget is created'''

        if title is None:
            if number_title:
                title_num = len(self.tabs) + 1
                title = f"{title_num} - {label}"
            else:
                raise ValueError('''The title of the tab has not been set.
                    Please set title - this must be a string, or set number_title to True.''')
        self.stack_list.addItem(title)

        if widget == 'form':
            widget = UIFormFactory.getQWidget(self)

        self.Stack.addWidget(widget)
        self.tabs[label] = widget
        self.num_tabs += 1
        height_multiplier = self.num_tabs + 1
        if self.layout_type != 'vertical':
            height_multiplier += 1
            self.stack_list.setMaximumWidth(
                self.stack_list.sizeHintForColumn(0) * 1.2 + self.list_margin_size * 2)
            self.stack_list.setMaximumHeight(
                self.stack_list.sizeHintForRow(0) * height_multiplier + self.list_margin_size * 2)
        self.stack_list.setMaximumHeight(self.stack_list.sizeHintForRow(0) * height_multiplier)

    def addTabs(self, labels):
        '''
        Adds multiple tabs to the StackedWidget
        All tabs contain an empty FormWidget and their title is
        their label, plus they are numbered according to their order.
        Parameters
        ----------
        labels: list of str
            The names that will be used to refer to the tabs.
        '''
        for label in labels:
            self.addTab(label)

    def getTab(self, label):
        ''' return the tab with label: label
        Parameter
        ---------
        label: str
            The label of the tab to be returned.
            Note, the label may be different to the tab's title.'''
        return self.tabs[label]

    def getTabs(self):
        ''' return the dict of tabs'''
        return self.tabs


class StackedWidget(QWidget, UIStackedWidget):
    def __init__(self, parent=None, layout='vertical'):
        # dockWidgetContents = QtWidgets.QWidget()

        QtWidgets.QWidget.__init__(self, parent)
        self.createStack(layout)


class StackedDockWidget(QtWidgets.QDockWidget):
    def __init__(self, parent=None, title=None, layout='vertical'):

        QtWidgets.QDockWidget.__init__(self, parent)
        widget = StackedWidget(parent, layout)
        self.setWidget(widget)
        if title is not None:
            self.setObjectName(title)
            self.setWindowTitle(title)

    def addTab(self, title, widget='form'):
        self.widget().addTab(title, widget)

    def addTabs(self, titles):
        self.widget().addTabs(titles)

    def getTab(self, label):
        return self.widget().getTab(label)

    def getTabs(self):
        return self.widget().getTabs()


class StackedWidgetFactory(QWidget):
    '''creates a StackedWidget with a list of the widgets' titles
    to the left hand side.
    The returned dockWidget must be added with
    main_window.addDockWidget(QtCore.Qt.RightDockWidgetArea, dockWidget)
    '''
    def getQDockWidget(parent=None, title=None, layout='vertical'):
        return StackedDockWidget(parent, title, layout)

    def getQWidget(parent=None, layout='vertical'):
        return StackedWidget(parent, layout)



import qdarkstyle
from PySide2.QtCore import QSettings, QThreadPool
from PySide2.QtGui import QKeySequence
from PySide2.QtWidgets import QAction, QMainWindow
from qdarkstyle.dark.palette import DarkPalette
from qdarkstyle.light.palette import LightPalette

from .ProgressTimerDialog import ProgressTimerDialog
from .SessionDialogs import AppSettingsDialog


class MainWindowWithProgressDialogs(QMainWindow):
    '''
    A base class for a main window, with a menu bar, and ability to
    create ProgressTimerDialogs.
    In a derived class's __init__ method, the __init__ method of this class
    should be called first, before any other initialisation.

    Properties of Note:
    -------------------
    self.settings
        This is a QSettings object that is used for
        settings such as the light/dark theme.
    self.progress_windows
        This is a dictionary of ProgressTimerDialog objects, where the key is
        the name of the progress window.
    '''
    def __init__(self, title, app_name, settings_name=None, organisation_name=None, **kwargs):

        super(MainWindowWithProgressDialogs, self).__init__(**kwargs)

        self.setWindowTitle(title)
        self.app_name = app_name
        self.threadpool = QThreadPool()

        if settings_name is None:
            settings_name = app_name
        if organisation_name is None:
            organisation_name = app_name

        self.settings = QSettings(organisation_name, settings_name)

        self.setAppStyle()

        self.createMenu()

        self._progress_windows = {}

    @property
    def progress_windows(self):
        return self._progress_windows

    # Create the menu ----------------------------------------------------------

    def createMenu(self):
        '''Create the menu bar, with the following options:

        File
            Exit
                This will exit the application without saving the current session

        Settings
            App Settings
                This will open a dialog to change the application settings, such as
                the light/dark theme

        Sets
        ----
        self.menu_bar: QMenuBar
            The menu bar
        self.menus: dict
            A dictionary of the menu names and the corresponding QMenu objects

        '''

        menu_bar = self.menuBar()

        # If we add the menus this way, then the menuBar takes ownership of the menus:
        file_menu = menu_bar.addMenu("File")

        exit_action = QAction("Exit", self)
        exit_action.setShortcut(QKeySequence.Quit)
        exit_action.triggered.connect(self.close)
        file_menu.addAction(exit_action)

        settings_menu = menu_bar.addMenu("Settings")

        app_settings_action = QAction("App Settings", self)
        app_settings_action.triggered.connect(self.createAppSettingsDialog)
        settings_menu.addAction(app_settings_action)

        menus = {"File": file_menu, "Settings": settings_menu}

        self.menu_bar = menu_bar
        self.menus = menus

    # Settings -----------------------------------------------------------------

    def setAppStyle(self):
        '''Sets app stylesheet, based on user's settings '''
        if self.settings.value("dark_mode") is None:
            self.settings.setValue("dark_mode", "true")
        if self.settings.value("dark_mode") == "true":
            style = qdarkstyle.load_stylesheet(palette=DarkPalette)
        else:
            style = qdarkstyle.load_stylesheet(palette=LightPalette)
        self.setStyleSheet(style)

    def createAppSettingsDialog(self):
        '''Create a dialog to change the application settings, such as the light/dark theme'''
        dialog = AppSettingsDialog(self)
        self.setAppSettingsDialogWidgets(dialog)
        dialog.Ok.clicked.connect(lambda: self.onAppSettingsDialogAccepted(dialog))
        dialog.Cancel.clicked.connect(dialog.close)
        dialog.open()

    def setAppSettingsDialogWidgets(self, dialog):
        '''Set the widgets on the app settings dialog, based on the
        current settings of the app'''
        if self.settings.value("dark_mode") == "true":
            dialog.widgets['dark_checkbox_field'].setChecked(True)
        else:
            dialog.widgets['dark_checkbox_field'].setChecked(False)

    def onAppSettingsDialogAccepted(self, dialog):
        '''This method is called when the user clicks
            "Ok" on the app settings dialog'''
        if dialog.widgets['dark_checkbox_field'].isChecked():
            self.settings.setValue("dark_mode", "true")
        else:
            self.settings.setValue("dark_mode", "false")
        self.setAppStyle()
        dialog.close()

    # Progress Bar -------------------------------------------------------------

    def createUnknownProgressWindow(self, process_name, title=None, detailed_text=None):
        '''
        Creates a progress bar with an unknown duration

        Instead of expecting to receive progress updates, this progress bar counts
        the number of seconds that have elapsed since it was created.

        This is useful for processes that take a long time, but for which we don't
        know how long they will take.

        Parameters
        ----------
        process_name : str
            The name of the process to be displayed in the progress bar
        title : str
            The title of the progress bar
        detailed_text : str
            The detailed text of the progress bar
        '''

        progress_window = ProgressTimerDialog(process_name, parent=self)
        self.saveReferenceToProgressWindow(progress_window, process_name)
        progress_window.show()

    def saveReferenceToProgressWindow(self, progress_window, process_name):
        self.progress_windows[process_name] = progress_window

    def finishProcess(self, process_name):
        '''
        Called when a process has finished.
        Closes the progress bar and sets the process_finished flag to True.
        '''
        self.process_finished = True
        self.progress_windows[process_name].close()



import sys

from PySide2 import QtWidgets

from eqt.ui import UIFormFactory


class MainUI(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        QtWidgets.QMainWindow.__init__(self, parent)

        pb = QtWidgets.QPushButton(self)
        pb.setText("Open Dialog with form layout")
        pb.clicked.connect(lambda: self.openFormDialog())

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(pb)
        widg = QtWidgets.QWidget()
        widg.setLayout(layout)

        self.setCentralWidget(widg)

        self.show()

    def openFormDialog(self):

        dialog = QtWidgets.QDialog(parent=self)
        bb = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok
                                        | QtWidgets.QDialogButtonBox.Cancel)

        bb.button(QtWidgets.QDialogButtonBox.Ok).clicked.connect(lambda: self.accepted())
        bb.button(QtWidgets.QDialogButtonBox.Cancel).clicked.connect(lambda: self.rejected())
        self.buttonBox = bb

        # create a form layout widget
        fw = UIFormFactory.getQWidget(parent=self)

        # ## Example on how to add elements to the form

        # add title
        qlabel = QtWidgets.QLabel(fw.groupBox)
        qlabel.setText("Form Widget")
        fw.addTitle(qlabel, 'title')

        # add input 1 as QLineEdit
        qlabel = QtWidgets.QLabel(fw.groupBox)
        qlabel.setText("Input 1: ")
        qwidget = QtWidgets.QLineEdit(fw.groupBox)
        qwidget.setClearButtonEnabled(True)
        # finally add to the form widget
        fw.addWidget(qwidget, qlabel, 'input1')

        # add separator
        fw.addSeparator('separator_1_2')

        # add input 2 as QComboBox
        qlabel = "Input2: "
        qwidget = QtWidgets.QComboBox(fw.groupBox)
        qwidget.addItem("option 1")
        qwidget.addItem("option 2")
        qwidget.setCurrentIndex(0)
        qwidget.setEnabled(True)
        # finally add to the form widget
        fw.addWidget(qwidget, qlabel, 'input2')

        # Example of using 'getWidget':
        fw.getWidget('input2', 'label').setText("Input 2: ")

        # add the button box to the vertical layout, but outside the
        # form layout
        fw.uiElements['verticalLayout'].addWidget(bb)
        dialog.setLayout(fw.uiElements['verticalLayout'])

        # store a reference
        self.fw = fw
        self.dialog = dialog

        dialog.exec()

    def accepted(self):
        print("accepted")
        print(self.fw.widgets['input1_field'].text())
        print(self.fw.widgets['input2_field'].currentText())

        self.dialog.close()

    def rejected(self):
        print("rejected")
        self.dialog.close()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)

    window = MainUI()

    sys.exit(app.exec_())



import json
import os
import shutil
import unittest
from datetime import datetime
from unittest import mock
from unittest.mock import patch

from PySide2.QtCore import QSettings, QThreadPool
from PySide2.QtWidgets import QMenu, QMenuBar

import eqt
from eqt.io import zip_directory
from eqt.ui.MainWindowWithSessionManagement import MainWindowWithSessionManagement

from . import skip_ci


@skip_ci
class TestMainWindowWithSessionManagementInit(unittest.TestCase):
    '''
    Tests the init method of the MainWindowWithSessionManagement class
    '''
    def setUp(self):
        # Testing the init method, so don't call init in the setUp
        self.title = "title"
        self.app_name = "App Name"

    def test_init_sets_title_and_app_name(self):
        smw = MainWindowWithSessionManagement(self.title, self.app_name)
        assert smw is not None
        self.assertEqual(smw.windowTitle(), "title")
        self.assertEqual(smw.app_name, "App Name")

    def test_init_creates_threadpool(self):
        smw = MainWindowWithSessionManagement(self.title, self.app_name)
        assert smw.threadpool is not None
        assert (isinstance(smw.threadpool, QThreadPool))

    def test_init_creates_settings(self):
        smw = MainWindowWithSessionManagement(self.title, self.app_name)
        assert smw.settings is not None
        assert (isinstance(smw.settings, QSettings))
        self.assertEqual(smw.settings.organizationName(), self.app_name)
        self.assertEqual(smw.settings.applicationName(), self.app_name)

    def test_init_creates_settings_when_settings_name_given(self):
        settings_name = "settings_name"
        smw = MainWindowWithSessionManagement(self.title, self.app_name, settings_name)
        assert smw.settings is not None
        assert (isinstance(smw.settings, QSettings))
        self.assertEqual(smw.settings.organizationName(), self.app_name)
        self.assertEqual(smw.settings.applicationName(), settings_name)

    @patch('eqt.ui.MainWindowWithSessionManagement.MainWindowWithSessionManagement.setAppStyle')
    def test_init_calls_setAppStyle(self, mock_setAppStyle):
        smw = MainWindowWithSessionManagement(self.title, self.app_name)
        smw.setAppStyle.assert_called_once()

    @patch('eqt.ui.MainWindowWithSessionManagement.MainWindowWithSessionManagement.createMenu',
           return_value=(1, {}))
    def test_init_calls_createMenu(self, mock_menu_bar):
        smw = MainWindowWithSessionManagement(self.title, self.app_name)
        smw.createMenu.assert_called_once()

    def test_sessions_directory_name_set(self):
        smw = MainWindowWithSessionManagement(self.title, self.app_name)
        self.assertEqual(smw.sessions_directory_name, "App-Name-Sessions")

    def test_init_initialises_vars(self):
        smw = MainWindowWithSessionManagement(self.title, self.app_name)
        self.assertEqual(smw.sessions_directory, None)
        self.assertEqual(smw.should_really_close, False)
        self.assertEqual(smw.progress_windows, {})

    @patch('eqt.ui.MainWindowWithSessionManagement.MainWindowWithSessionManagement.setupSession')
    def test_init_calls_setUpSession(self, mock_setupSession):
        smw = MainWindowWithSessionManagement(self.title, self.app_name)
        smw.setupSession.assert_called_once()


@skip_ci
class TestMainWindowWithSessionManagementMenuBar(unittest.TestCase):
    '''
    Tests the expected menu bar is created
    '''
    def setUp(self):
        self.title = "title"
        self.app_name = "app_name"
        self.smw = MainWindowWithSessionManagement(self.title, self.app_name)

    def test_createMenu_sets_menu_bar_and_menus(self):
        # first remove the menu bar and menus which are created in the init
        del self.smw.menu_bar
        del self.smw.menus
        self.smw.createMenu()
        assert hasattr(self.smw, "menu_bar")
        assert hasattr(self.smw, "menus")
        assert isinstance(self.smw.menu_bar, QMenuBar)
        assert isinstance(self.smw.menus, dict)
        # dict should contain the expected menus
        assert "File" in self.smw.menus
        assert "Settings" in self.smw.menus
        assert isinstance(self.smw.menus["File"], QMenu)
        assert isinstance(self.smw.menus["Settings"], QMenu)

    def test_menu_has_file_and_settings_menu(self):
        actions = self.smw.menu_bar.actions()
        assert actions[0].text() == "File"
        assert actions[1].text() == "Settings"

    def test_file_menu_has_expected_actions(self):
        menus = self.smw.menu_bar.findChildren(QMenu)
        file_menu = menus[0]
        assert file_menu.actions()[0].text() == "Save"
        assert file_menu.actions()[1].text() == "Save + Exit"
        assert file_menu.actions()[2].text() == "Exit"

    def test_settings_menu_has_expected_actions(self):
        menus = self.smw.menu_bar.findChildren(QMenu)
        settings_menu = menus[1]
        self.assertEqual(settings_menu.actions()[0].text(), "App Settings")
        self.assertEqual(settings_menu.actions()[1].text(), "Set Session Directory")


@skip_ci
class TestMainWindowWithSessionManagementSetupSession(unittest.TestCase):
    '''
    Tests the setupSession method of the MainWindowWithSessionManagement class

    This method is responsible for setting up the session directory and session selector
    '''
    def setUp(self):
        self.title = "title"
        self.app_name = "app_name"
        self.smw = MainWindowWithSessionManagement(self.title, self.app_name)

    def test_setupSession_when_sessions_folder_setting_is_None(self):
        self.smw.settings = mock.MagicMock()
        self.smw.createSessionsDirectorySelectionDialog = mock.MagicMock()
        self.smw.createSessionSelector = mock.MagicMock()
        self.smw.settings.value = mock.MagicMock(return_value=None)
        self.smw.setupSession()
        self.smw.createSessionsDirectorySelectionDialog.assert_called_once_with(new_session=True)
        self.smw.createSessionSelector.assert_not_called()

    def test_setupSession_when_sessions_folder_setting_is_not_None(self):
        self.smw.settings = mock.MagicMock()
        self.smw.createSessionsDirectorySelectionDialog = mock.MagicMock()
        self.smw.createSessionSelector = mock.MagicMock()
        session_folder_name = "session_folder_name"
        os.mkdir(session_folder_name)
        try:
            self.smw.settings.value = mock.MagicMock(return_value=session_folder_name)
            self.smw.setupSession()
            self.smw.createSessionsDirectorySelectionDialog.assert_not_called()
            self.smw.createSessionSelector.assert_called_once()
            self.assertEqual(self.smw.sessions_directory, session_folder_name)

        except Exception as e:
            raise e

        finally:
            if os.path.basename(os.getcwd()) == session_folder_name:
                os.chdir("..")
                os.rmdir(session_folder_name)

            else:
                os.rmdir(session_folder_name)

    def test_setupSession_when_given_nonexistent_sessions_folder(self):
        self.smw.settings = mock.MagicMock()
        self.smw.createSessionsDirectorySelectionDialog = mock.MagicMock()
        self.smw.createSessionSelector = mock.MagicMock()
        session_folder_name = "session_folder_name"
        self.smw.settings.value = mock.MagicMock(return_value=session_folder_name)
        self.smw.setupSession()
        self.smw.createSessionsDirectorySelectionDialog.assert_called_once_with(new_session=True)
        self.smw.createSessionSelector.assert_not_called()


@skip_ci
class TestMainWindowWithSessionManagementCreateSessionSelector(unittest.TestCase):
    '''
    Tests the createSessionSelector method of the MainWindowWithSessionManagement class

    This method is responsible for creating the session selector if any sessions exist,
    or calling a method to create a new session if no sessions exist
    '''
    def setUp(self):
        self.title = "title"
        self.app_name = "app_name"
        self.smw = MainWindowWithSessionManagement(self.title, self.app_name)
        os.mkdir("Test Folder")
        os.chdir("Test Folder")
        os.mkdir("Session Folder")
        self.smw.sessions_directory = "."

    def test_createSessionSelector_when_no_session_zips_exist(self):
        self.smw.loadSessionNew = mock.MagicMock()
        eqt.ui.SessionDialogs.LoadSessionDialog = mock.MagicMock()
        self.smw.createSessionSelector()
        self.smw.loadSessionNew.assert_called_once()
        eqt.ui.SessionDialogs.LoadSessionDialog.assert_not_called()

    def test_createSessionSelector_when_session_zips_exist(self):
        # Make 2 zip files in the sessions directory:
        os.chdir("Session Folder")
        shutil.make_archive("_session1_08-02-22", "zip")
        shutil.make_archive("session_2_08-02-22", "zip")
        os.chdir("..")

        zip_folders = ["_session1 08-02-22", "session_2 08-02-22"]

        self.smw.createLoadSessionDialog = mock.MagicMock()
        self.smw.loadSessionNew = mock.MagicMock()
        self.smw.createSessionSelector()
        self.smw.loadSessionNew.assert_not_called()

        try:
            self.smw.createLoadSessionDialog.assert_called_once_with(zip_folders)
        except AssertionError:
            # zip files could be returned in reverse order
            self.smw.createLoadSessionDialog.assert_called_once_with(zip_folders[::-1])

    def tearDown(self):
        os.chdir("..")
        shutil.rmtree("Test Folder")


@skip_ci
class TestSelectLoadSessionsDirectorySelectedInSessionSelector(unittest.TestCase):
    '''
    Tests the `selectLoadSessionsDirectorySelectedInSessionSelector` method of the
    `MainWindowWithSessionManagement` class

    This method should close the passed dialog, and call
    `createSessionsDirectorySelectionDialog(new_session=True)`
    '''
    def setUp(self):
        self.title = "title"
        self.app_name = "app_name"
        self.smw = MainWindowWithSessionManagement(self.title, self.app_name)
        self.load_session_dialog = mock.MagicMock()
        self.load_session_dialog.close = mock.MagicMock()
        self.smw.createSessionsDirectorySelectionDialog = mock.MagicMock()

    def test_selectLoadSessionsDirectorySelectedInSessionSelector(self):
        self.smw.selectLoadSessionsDirectorySelectedInSessionSelector(self.load_session_dialog)
        assert self.load_session_dialog.close.called_once()
        assert self.smw.createSessionsDirectorySelectionDialog.called_once_with(new_session=True)


@skip_ci
class TestCreateSessionFolder(unittest.TestCase):
    '''
    Tests the createSessionFolder method of the MainWindowWithSessionManagement class

    This method is responsible for creating a folder for the session and moving into it
    '''
    def setUp(self):
        self.title = "title"
        self.app_name = "app_name"

        self.smw = MainWindowWithSessionManagement(self.title, self.app_name)

        os.mkdir("Test Folder")
        os.chdir("Test Folder")
        self.date_time = datetime.now().strftime("%d-%m-%Y-%H-%M-%S")

    def test_createSessionFolder(self):
        session_folder_name = "app_name-" + self.date_time
        # We need to mock the sessions_directory method, as it has not been set up yet
        # We want the session folder to be made in the current directory for this test
        self.smw.sessions_directory = "."
        self.smw.createSessionFolder()
        assert os.path.exists(session_folder_name)

    def tearDown(self):
        os.chdir("..")
        shutil.rmtree("Test Folder")


@skip_ci
class TestLoadSessionConfig(unittest.TestCase):
    '''
    Tests the loadSessionConfig method of the `MainWindowWithSessionManagement` class.

    This method is responsible for unzipping a session folder and saving the contents
    of the `.json` session file to `MainWindowWithSessionManagement.config`.
    '''
    def setUp(self):
        '''
        Create a session zip file, which contains a session.json file'''
        os.mkdir("Test Folder")
        os.chdir("Test Folder")

        self.title = "title"
        self.app_name = "app_name"
        self.smw = MainWindowWithSessionManagement(self.title, self.app_name)
        self.smw.sessions_directory = "Sessions Folder"
        self.session_folder = os.path.join(self.smw.sessions_directory,
                                           "app name 01-01-2020-00-00-00")

        self.config = {
            'test_key': 'test_value', 'test_key2': 'test_value2', 'test_int_key': 1,
            'test_float_key': 1.0, 'test_bool_key': True, 'test_list_key': [1, 2, 3],
            'test_dict_key': {'test_key': 'test_value'}}

        os.mkdir(self.smw.sessions_directory)
        os.mkdir(self.session_folder)

        session_file = os.path.join(self.session_folder, "session.json")
        with open(session_file, "w+") as f:
            json.dump(self.config, f)

    def test_loadSessionConfig(self):
        zip_directory(self.session_folder, compress=False)
        self.smw.loadSessionConfig(self.session_folder)
        self.assertEqual(self.config, self.smw.config)

    def tearDown(self):
        os.chdir("..")
        try:
            shutil.rmtree("Test Folder")
        except Exception:
            os.chdir("..")
            shutil.rmtree("Test Folder")


@skip_ci
class TestSaveSession(unittest.TestCase):
    '''
    Tests:
    - saveSession method of the MainWindowWithSessionManagement class
    - moveSessionFolder method of the MainWindowWithSessionManagement class
    - saveSessionConfigToJson method of the MainWindowWithSessionManagement class

    '''
    def setUp(self):
        self.title = "title"
        self.app_name = "app_name"
        self.smw = MainWindowWithSessionManagement(self.title, self.app_name)
        self.session_name = self.app_name

    @mock.patch('eqt.ui.MainWindowWithSessionManagement.zip_directory')
    def test_saveSession(self, mock_zip_directory):
        self.smw.moveSessionFolder = mock.MagicMock()
        self.smw.saveSessionConfigToJson = mock.MagicMock()
        self.smw.current_session_folder = mock.MagicMock()
        self.smw.saveSession(self.session_name, compress=False)
        self.smw.moveSessionFolder.assert_called_once_with(self.session_name)
        self.smw.saveSessionConfigToJson.assert_called_once()
        mock_zip_directory.assert_called_once_with(self.smw.current_session_folder, False)

    def test_moveSessionFolder(self):
        os.mkdir("Test Folder")
        new_folder_to_save_to = os.path.join(
            "Test Folder", self.session_name + "_" + datetime.now().strftime("%d-%m-%Y-%H-%M"))

        # Make the current session folder, with a test file inside it:

        self.smw.sessions_directory = os.path.abspath("Test Folder")
        current_session_folder = os.path.join(self.smw.sessions_directory,
                                              "Current Session Folder")
        os.mkdir(current_session_folder)
        # Write file inside the folder:
        with open(os.path.join(current_session_folder, "test_file.txt"), "w+") as f:
            f.write("test")
        # os.chdir("Current Session Folder")
        self.smw.current_session_folder = current_session_folder

        try:
            self.smw.moveSessionFolder(self.session_name)
            self.assertEqual(os.path.abspath(self.smw.current_session_folder),
                             os.path.abspath(new_folder_to_save_to))
            self.assertTrue(os.path.exists(new_folder_to_save_to))
            self.assertTrue(os.path.exists(os.path.join(new_folder_to_save_to, "test_file.txt")))
            self.assertFalse(os.path.exists("Current Session Folder"))
            self.assertFalse(
                os.path.exists(os.path.join("Current Session Folder", "test_file.txt")))

        except Exception as e:
            raise e

        finally:
            shutil.rmtree("Test Folder")

    def test_saveSessionConfigToJson(self):

        self.config = {
            'test_key': 'test_value', 'test_key2': 'test_value2', 'test_int_key': 1,
            'test_float_key': 1.0, 'test_bool_key': True, 'test_list_key': [1, 2, 3],
            'test_dict_key': {'test_key': 'test_value'}}

        config = {}
        config.update(self.config)

        # datetime gets added in the saveSessionConfigToJson method:
        self.config['datetime'] = datetime.now().strftime("%d-%m-%Y-%H-%M")

        self.smw.getSessionConfig = mock.MagicMock(return_value=config)

        os.mkdir("Test Folder")
        os.chdir("Test Folder")
        self.smw.current_session_folder = "."
        try:
            self.smw.saveSessionConfigToJson()
            session_file = os.path.join(os.getcwd(), "session.json")
            with open(session_file, "r") as f:
                config = json.load(f)
            self.assertEqual(config, self.config)
        except Exception as e:
            raise e
        finally:
            os.chdir("..")
            shutil.rmtree("Test Folder")


@skip_ci
class TestRemoveTempMethods(unittest.TestCase):
    '''
    Tests:
    - removeTempAndClose method of the MainWindowWithSessionManagement class
    - removeTemp method of the MainWindowWithSessionManagement class
    '''
    def setUp(self):
        self.title = "title"
        self.app_name = "app_name"
        self.smw = MainWindowWithSessionManagement(self.title, self.app_name)

    def test_removeTemp_when_current_session_folder_exists(self):
        try:
            self.smw.current_session_folder = "Test Session Folder"
            os.mkdir(self.smw.current_session_folder)
            self.smw.removeTemp()
            self.assertFalse(os.path.exists(self.smw.current_session_folder))
        except Exception as e:
            try:
                shutil.rmtree(self.smw.current_session_folder)
            except Exception:
                raise e

    def test_removeTemp_when_current_session_folder_does_not_exist(self):
        self.smw.current_session_folder = "Test Session Folder"
        # Test this raises a FileNotFoundError:
        with self.assertRaises(FileNotFoundError):
            self.smw.removeTemp()

    def test_removeTempAndClose(self):
        process_name = "Test Process"
        self.smw.removeTemp = mock.MagicMock()
        self.smw.close = mock.MagicMock()
        self.smw.finishProcess = mock.MagicMock()
        self.smw.should_really_close = False
        self.smw.removeTempAndClose(process_name)
        self.smw.removeTemp.assert_called_once()
        self.smw.finishProcess.assert_called_once_with(process_name)
        self.smw.close.assert_called_once()



from PySide2 import QtCore
from PySide2.QtWidgets import QSlider


class UISliderWidget(QSlider):
    '''Creates a Slider widget which updates
    a QLabel with its value (which may be scaled
    to a non-integer value by setting the scale_factor)'''
    def __init__(self, label, scale_factor=1):
        QSlider.__init__(self)
        self.label = label
        self.scale_factor = scale_factor
        self.setOrientation(QtCore.Qt.Horizontal)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setTickPosition(QSlider.TicksBelow)
        self.valueChanged.connect(self.show_slider_value)

    def get_slider_value(self):
        return self.value() * self.scale_factor

    def show_slider_value(self):
        value = self.get_slider_value()
        self.label.setText(str(value))



import sys

from PySide2 import QtWidgets

from eqt.ui import FormDialog
from eqt.ui.UISliderWidget import UISliderWidget


class MainUI(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        QtWidgets.QMainWindow.__init__(self, parent)

        pb = QtWidgets.QPushButton(self)
        pb.setText("Open Dialog with form layout")
        pb.clicked.connect(lambda: self.openFormDialog())

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(pb)
        widg = QtWidgets.QWidget()
        widg.setLayout(layout)

        self.setCentralWidget(widg)

        # create dialog to be opened later:
        dialog = FormDialog(parent=self, title='Example')

        # ## Example on how to add elements to the
        dialog.addWidget(QtWidgets.QLabel('test label'), 'Label: ', 'label')
        dialog.addWidget(QtWidgets.QCheckBox('test checkbox'), 'CheckBox: ', 'checkBox')
        combobox = QtWidgets.QComboBox()
        combobox.addItems(['test1', 'test2'])
        dialog.addWidget(combobox, 'ComboBox: ', 'comboBox')
        dialog.addWidget(QtWidgets.QDoubleSpinBox(), 'DoubleSpinBox: ', 'doubleSpinBox')
        dialog.addWidget(QtWidgets.QSpinBox(), 'SpinBox: ', 'spinBox')
        dialog.addWidget(QtWidgets.QSlider(), 'Slider: ', 'slider')
        dialog.addWidget(UISliderWidget(QtWidgets.QLabel()), 'UISliderWidget: ', 'uiSliderWidget')
        dialog.addWidget(QtWidgets.QRadioButton('test 1'), 'RadioButton 1: ', 'radioButton')
        dialog.addWidget(QtWidgets.QRadioButton('test 2'), 'RadioButton 2: ', 'radioButton')
        dialog.addWidget(QtWidgets.QTextEdit('test'), 'TextEdit: ', 'textEdit')
        dialog.addWidget(QtWidgets.QPlainTextEdit('test'), 'PlainTextEdit: ', 'plainTextEdit')
        dialog.addWidget(QtWidgets.QLineEdit('test'), 'LineEdit: ', 'lineEdit')
        button = QtWidgets.QPushButton('test')
        button.setCheckable(True)
        dialog.addWidget(button, 'Button: ', 'button')

        # store a reference
        self.dialog = dialog

        self.show()

    def openFormDialog(self):
        self.dialog.exec()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)

    window = MainUI()

    sys.exit(app.exec_())



import sys

from PySide2 import QtWidgets

from eqt.ui import FormDialog, UIFormWidget


class MainUI(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        QtWidgets.QMainWindow.__init__(self, parent)

        # dialog form
        self.dialog = FormDialog(parent=self, title='Form Dialog example insert widget')
        self.addWidgetsToExampleForm(self.dialog)
        buttoninsertvertical = QtWidgets.QPushButton()
        buttoninsertvertical.setText("Insert widget in vertical layout")
        self.dialog.addSpanningWidget(buttoninsertvertical, 'Button insert vertical')
        buttoninsertvertical.clicked.connect(lambda: self.insert_vertical())

        # create a FormDockWidget
        dock = UIFormWidget.FormDockWidget(parent=self)
        dock.setWindowTitle('Dock Widget Example insert widget')
        self.addWidgetsToExampleForm(dock)

        # create button for Form Dialog
        pb = QtWidgets.QPushButton(self)
        pb.setText("Open Form Dialog")
        pb.clicked.connect(self.openFormDialog)

        # create window layout
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(pb)
        layout.addWidget(dock)
        widg = QtWidgets.QWidget()
        widg.setLayout(layout)
        self.setCentralWidget(widg)

        self.dialog.onCancel = self.onCancel
        self.show()

    def openFormDialog(self):
        self.dialog.open()

    def addWidgetsToExampleForm(self, form):
        form.addWidget(QtWidgets.QLineEdit(), "Initial widget row 0: ", 'Initial widget row 0')
        form.addSpanningWidget(QtWidgets.QLabel("Initial widget row 1"), 'Initial widget row 1')
        # add QComboBox
        qwidget = QtWidgets.QComboBox(form)
        qwidget.addItem("0")
        qwidget.addItem("1")
        form.addWidget(qwidget, "Initial widget row 2", 'Initial widget row 2')
        buttoninsert = QtWidgets.QPushButton()
        buttoninsert.setText("Insert widgets")
        form.addSpanningWidget(buttoninsert, 'Button insert widgets')
        buttoninsert.clicked.connect(lambda: self.insert_form(form, buttoninsert))

    def insert_vertical(self):
        self.dialog.insertWidgetToVerticalLayout(
            1, QtWidgets.QPushButton("Inserted widget in vertical layout"))
        print(
            "\nThe dictionary of widgets does not change after insertion in the vertical layout.")
        self.dialog.getWidget('Button insert vertical').setEnabled(False)

    def insert_form(self, form, button):
        qlabel = QtWidgets.QLabel(form)
        qlabel.setText("Widget inserted in row 0: ")
        qwidget = QtWidgets.QLineEdit(form)
        form.insertWidget(0, 'inserted widget', qwidget, qlabel)
        buttonspanning = QtWidgets.QPushButton(self)
        buttonspanning.setText("Spanning widget inserted in row 2")
        form.insertWidget(2, 'inserted spanning widget', buttonspanning)
        print('\nDictionary of widgets after insertion in the form layout:\n' +
              str(form.getWidgets()))
        button.setEnabled(False)

    def onCancel(self):
        if not hasattr(self.dialog.formWidget, 'widget_states'):
            if self.dialog.getWidget('Button insert vertical').isEnabled() is False:
                self.dialog.removeWidgetFromVerticalLayout(
                    self.dialog.getWidgetFromVerticalLayout(1))

            if self.dialog.getWidget('Button insert widgets').isEnabled() is False:
                self.dialog.formWidget._popWidget(self.dialog.formWidget.default_widget_states,
                                                  'inserted widget')
                self.dialog.formWidget._popWidget(self.dialog.formWidget.default_widget_states,
                                                  'inserted spanning widget')
                self.dialog.formWidget.removeWidget('inserted widget')
                self.dialog.formWidget.removeWidget('inserted spanning widget')
        else:
            if self.dialog.getWidget('Button insert vertical').isEnabled(
            ) != self.dialog.getWidgetStates()['Button insert vertical_field']['enabled'] is True:
                self.dialog.removeWidgetFromVerticalLayout(
                    self.dialog.getWidgetFromVerticalLayout(1))
            if self.dialog.getWidget('Button insert widgets').isEnabled(
            ) != self.dialog.getWidgetStates()['Button insert widgets_field']['enabled'] is True:
                self.dialog.formWidget.removeWidget('inserted widget')
                self.dialog.formWidget.removeWidget('inserted spanning widget')


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)

    window = MainUI()

    sys.exit(app.exec_())



import sys

from PySide2 import QtWidgets
from PySide2.QtWidgets import QApplication

from eqt import __version__
from eqt.ui.MainWindowWithSessionManagement import MainWindowWithSessionManagement
from eqt.ui.UIFormWidget import FormWidget
from eqt.ui.UISliderWidget import UISliderWidget


class ExampleMainWindowWithSessionManagement(MainWindowWithSessionManagement):
    '''
    This is an example of a MainWindowWithSessionManagement.
    It is used to show how to use the MainWindowWithSessionManagement class.

    To test out this example, change the state of the widgets and save the session.
    Then load the session again and check if the state of the widgets is the same as before.
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        form_widget = FormWidget(self)
        add_every_widget_to_form(form_widget)
        self.setCentralWidget(form_widget)
        self.addToMenu()

    def addToMenu(self):
        ''' Adds an example extra menu.'''
        extra_menu = self.menu_bar.addMenu("Extra Menu")
        self.menu_bar.addMenu(extra_menu)
        self.menus["Extra"] = extra_menu

    def getSessionConfig(self):
        '''
        This function is called when the user wants to save the current session.
        We need to add the state of the widgets to the config.

        Returns
        -------
        config : dict
            The config of the current session.
        '''

        form_widget = self.centralWidget()
        all_widget_states = form_widget.getAllWidgetStates()

        config = {'form_widget': all_widget_states}

        return config

    def finishLoadConfig(self, process_name):
        '''
        This function is called after the config of the chosen session is loaded.
        It is used to set the state of the widgets.
        '''
        form_widget_config = self.config['form_widget']
        form_widget = self.centralWidget()
        form_widget.applyWidgetStates(form_widget_config)
        super().finishLoadConfig(process_name)


def add_every_widget_to_form(form):
    '''
    Generate every widget and add it to the form

    Parameters
    ----------
    form : FormWidget, FormDialog or FormDockWidget
        The form to add the widgets to
    '''
    form.addWidget(QtWidgets.QLabel('test label'), 'Label: ', 'label')
    form.addWidget(QtWidgets.QCheckBox('test checkbox'), 'CheckBox: ', 'checkBox')
    form.addWidget(QtWidgets.QComboBox(), 'ComboBox: ', 'comboBox')
    form.addWidget(QtWidgets.QDoubleSpinBox(), 'DoubleSpinBox: ', 'doubleSpinBox')
    form.addWidget(QtWidgets.QSpinBox(), 'SpinBox: ', 'spinBox')
    form.addWidget(QtWidgets.QSlider(), 'Slider: ', 'slider')
    form.addWidget(UISliderWidget(QtWidgets.QLabel()), 'UISliderWidget: ', 'uiSliderWidget')
    form.addWidget(QtWidgets.QRadioButton('test'), 'RadioButton: ', 'radioButton')
    form.addWidget(QtWidgets.QTextEdit('test'), 'TextEdit: ', 'textEdit')
    form.addWidget(QtWidgets.QPlainTextEdit('test'), 'PlainTextEdit: ', 'plainTextEdit')
    form.addWidget(QtWidgets.QLineEdit('test'), 'LineEdit: ', 'lineEdit')
    form.addWidget(QtWidgets.QPushButton('test'), 'Button: ', 'button')


def create_main_window():
    window = ExampleMainWindowWithSessionManagement(f"Example Session Main Window{__version__}",
                                                    "Example0", settings_name="Example0")

    return window


def main():
    # open main window:
    app = QApplication(sys.argv)
    window = create_main_window()
    window.show()
    app.exec_()


if __name__ == "__main__":
    main()



from PySide2 import QtCore, QtWidgets

from . import UIFormFactory


class FormDialog(QtWidgets.QDialog):
    def __init__(self, parent=None, title=None):
        super().__init__(parent)
        self.setWindowFlag(QtCore.Qt.WindowContextHelpButtonHint, False)

        self.buttonBox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok
                                                    | QtWidgets.QDialogButtonBox.Cancel)
        self.formWidget = UIFormFactory.getQWidget(parent=self)
        # set the layout of the dialog
        self.setLayout(self.formWidget.uiElements['verticalLayout'])

        if title is not None:
            self.setWindowTitle(title)
        # add button box to the UI
        self.formWidget.uiElements['verticalLayout'].addWidget(self.buttonBox)
        self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).clicked.connect(self._onOk)
        self.buttonBox.button(QtWidgets.QDialogButtonBox.Cancel).clicked.connect(self._onCancel)

    @property
    def Ok(self):
        '''Returns a reference to the Dialog Ok button to connect its signals.'''
        return self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok)

    @property
    def Cancel(self):
        '''Returns a reference to the Dialog Cancel button to connect its signals.'''
        return self.buttonBox.button(QtWidgets.QDialogButtonBox.Cancel)

    def _onOk(self):
        '''Saves the widget states and calls `onOk`.'''
        self.saveAllWidgetStates()
        self.onOk()
        self.close()

    def _onCancel(self):
        '''Calls `onCancel`, closes the FormDialog and restores the previously saved states
        or the default states.'''
        self.onCancel()
        self.close()
        self.restoreAllSavedWidgetStates()

    def onOk(self):
        '''Called when the dialog's "Ok" button is clicked.
        Can be redefined to add additional functionality on "Ok".'''
        pass

    def onCancel(self):
        '''Called when the dialog's "Cancel" button is clicked.
        Can be redefined to add additional functionality on "Cancel"'''
        pass

    @property
    def widgets(self):
        return self.formWidget.widgets

    @property
    def groupBox(self):
        return self.formWidget.groupBox

    def addWidget(self, qwidget, qlabel=None, name=None, layout='form'):
        '''
        Adds a widget to the layout. In particular, adds a qwidget and a qlabel widget
        in the same row of the form layout if layout is 'form' and adds a spanning widget
        to the vertical layout if layout is 'vertical'.

        Parameters
        ----------
        qwidget : QWidget
        qlabel : qlabel widget or str
            only supported when layout is 'form'
        name : str
            only supported when layout is 'form'
        layout : 'form' or 'vertical'
                'form' - adds to the `groupBoxFormLayout`,
                'vertical' - adds to the `verticalLayout` below the form.
        '''
        if layout == 'vertical':
            if name is not None or qlabel is not None:
                raise ValueError('`qlabel` and `name` are unsupported when `layout=vertical`')
            self.formWidget.uiElements['verticalLayout'].addWidget(qwidget)
        elif layout == 'form':
            if name is None:
                raise ValueError('To add the widget to the form, please set name.')
            if qlabel is None:
                raise ValueError('To add the widget to the form, please set label.')
            self.formWidget.addWidget(qwidget, qlabel, name)
        else:
            raise ValueError(f"layout '{layout}' unrecognised: expected 'form' or 'vertical'")

    def addSpanningWidget(self, qwidget, name=None, layout='form'):
        '''
        Adds a spanning widget occupying the full row in the layout.

        Parameters
        ----------
        qwidget : QWidget
        name : str
            only supported when layout is 'form'
        layout : 'form' or 'vertical'
                'form' - adds to the `groupBoxFormLayout`,
                'vertical' - adds to the `verticalLayout` below the form.
        '''
        if layout == 'vertical':
            if name is not None:
                raise ValueError(f"`name='{name}'` cannot be given if `layout='{layout}'`")
            self.formWidget.uiElements['verticalLayout'].addWidget(qwidget)
        elif layout == 'form':
            if name is None:
                raise ValueError('To add the widget to the form, please set name')
            self.formWidget.addSpanningWidget(qwidget, name)
        else:
            raise ValueError(
                f"layout {layout} is not recognised, must be set to 'form' or 'vertical'")

    def insertWidget(self, row, name, qwidget, qlabel=None):
        '''
        Inserts a labelled widget, or a spanning widget, to the form layout.
        The position in the form is specified by row. If row is out of bounds, the widget
        is added at the end of the form. An error is raised if `name` is already in use
        by another widget in the form. The entries associated with the widget are added
        to the widget dictionary and the default-widget-states
        dictionary.

        Parameters
        ----------
        row : int
            The position in the form where the widget is added.
        name : str
            The string associated to the qwidget and qlabel.
        qwidget : QWidget
            The widget to be added on the right hand side of the form or as a spanning widget.
        qlabel : qlabel widget or str
            The qlabel widget, or a str from which a qlabel widget is created, to be added
            on the left hand side of the form. If qlabel is `None` the widget spans the full
            width of the form.
        '''
        self.formWidget.insertWidget(row, name, qwidget, qlabel)

    def insertWidgetToVerticalLayout(self, row, qwidget):
        '''Inserts a widget to the vertical layout at position specified by row.'''
        self.formWidget.uiElements['verticalLayout'].insertWidget(row, qwidget)

    def getWidgetFromVerticalLayout(self, index):
        '''Returns the widget in the vertical layout located at position index.'''
        return self.formWidget.uiElements['verticalLayout'].itemAt(index).widget()

    def getIndexFromVerticalLayout(self, widget):
        '''
        Returns the index of the widget in the vertical layout.

        Parameters
        -------------
        widget : QWidget
            The widget in the layout.

        Return
        ------------
        int
            The index of the widget in the layout.
        '''
        return self.formWidget.uiElements['verticalLayout'].indexOf(widget)

    def removeWidget(self, name):
        '''
        Removes the widget with the specified name from the form layout.
        This method deletes the qwidget, and qlabel if present, from the widgets dictionary
        and sets their parent to `None`.

        Parameters
        ----------
        name : str
            The name of the widget to be removed.

        Returns
        -------
        tuple or QWidget
            If the widget has a corresponding label, a tuple containing the widget
            and label is returned. Otherwise, only the widget is returned.
        '''
        return self.formWidget.removeWidget(name)

    def removeWidgetFromVerticalLayout(self, widget):
        '''Removes a widget from the vertical layout.

        Parameters
        ----------
        widget : QWidget
            The widget to be removed.
        '''
        self.formWidget.uiElements['verticalLayout'].removeWidget(widget)
        widget.setParent(None)
        return widget

    def getNumWidgets(self):
        '''Returns the number of widgets in the form.'''
        return self.formWidget.getNumWidgets()

    def getWidget(self, name, role='field'):
        '''
        Returns the widget by the name with which it has been added.
        By default it returns the widget that is the field in the form.
        The user can get the label by specifying the role to be label.
        Raises ValueError if the role is not field or label.
        '''
        return self.formWidget.getWidget(name, role)

    def getWidgets(self):
        '''Returns a dictionary of the widgets currently present in the form.'''
        return self.formWidget.getWidgets()

    def getWidgetRow(self, name, role='field'):
        '''
        Returns the widget row in the form layout by the widget name.
        This is the row of the widget in the form layout.
        '''
        return self.formWidget.getWidgetRow(name, role)

    def setWidgetVisible(self, name, visible):
        '''
        Sets the visibility of the widget and associated label with the given name.

        Parameters
        ----------
        name : str
            The name of the widget to set visible/invisible
        visible : bool
            True to set the widget visible, False to hide it
        '''
        self.formWidget.setWidgetVisible(name, visible)

    def saveAllWidgetStates(self):
        '''
        Saves the state of all widgets currently present in the form.
        To later restore the states, use `restoreAllSavedWidgetStates()`.
        '''
        self.formWidget.saveAllWidgetStates()

    def getWidgetStates(self):
        '''Deprecated. Use `getSavedWidgetStates`.'''
        return self.formWidget.getWidgetStates()

    def getSavedWidgetStates(self):
        '''Returns the saved widget states.'''
        return self.formWidget.getSavedWidgetStates()

    def getDefaultWidgetStates(self):
        '''Returns the saved default widget states.'''
        return self.formWidget.getDefaultWidgetStates()

    def restoreAllSavedWidgetStates(self):
        '''
        Restore all widgets in the form to the state saved by `saveAllWidgetStates()`.
        If `saveAllWidgetStates()` method was not previously invoked, do nothing.
        '''
        self.formWidget.restoreAllSavedWidgetStates()

    def getAllWidgetStates(self):
        '''
        Returns
        -------
        dict
            Format: {'widget_name': {'value': str | bool | int, 'enabled': bool, 'visible': bool,
            'widget_row': int}, ...}.
            e.g. {'widget1': {'value': 1, 'enabled': True, 'visible': True, 'widget_row': 0},
            'widget2': {'value': 2, 'enabled': False, 'visible': False, 'widget_row': 1}}.
        '''
        return self.formWidget.getAllWidgetStates()

    def getWidgetState(self, widget, role=None):
        '''
        Returns the current state of the widget in the form.

        Parameters
        ----------
        widget : QWidget or str
            The widget or its name (or its name + '_field' or '_label', when role is None) to get
            the state of.
        role : str, optional, default None, values: 'label', 'field', None.
            The role of the widget to apply the state to. This is used only if `widget` is the
            widget name string.

        Returns
        -------
        state : dict
            Format: {'value': str | bool | int, 'enabled': bool, 'visible': bool,
            'widget_row' : int}.
            e.g. {'value': 1, 'enabled': True, 'visible': True, 'widget_row' : 0}.
            This can be used to restore the state of the widget using `setWidgetState()`.
        '''
        return self.formWidget.getWidgetState(widget, role)

    def applyWidgetState(self, name, state, role=None):
        '''
        Applies the given `state` to the widget associated with `name` and `role`.
        If role is None, the role is assigned to be 'field'.

        Parameters
        ----------
        name : str
            The name of the widget to apply the state to.
        role : str, optional, default None, values: 'label', 'field', None.
            The role of the widget to apply the state to.
        state : dict
            Format: {'value': str | bool | int, 'enabled': bool, 'visible': bool,
            'widget_row' : int}.
            e.g. {'value': 1, 'enabled': True, 'visible': True, 'widget_row' : 0}.
        '''
        return self.formWidget.applyWidgetState(name, state, role)

    def applyWidgetStates(self, states):
        '''
        Applies the given states to the form's widgets. It raises an error if the keys in the
        dicitonary of states and the keys in the dictionary of widgets in the form are not the
        same.

        Parameters
        ----------
        states : dict
            Format: {'name_field': {'value': str | bool | int, 'enabled': bool, 'visible': bool,
            'widget_row' : int}, 'name_label': {'value': str | bool | int, 'enabled': bool,
            'visible': bool, 'widget_row' : int}, ...}.
            e.g. {'widget1': {'value': 1, 'enabled': True, 'visible': True, 'widget_row': 0},
                  'widget2': {'value': 2, 'enabled': False, 'visible': False, 'widget_row': 1}}.
        '''
        return self.formWidget.applyWidgetStates(states)


class AdvancedFormDialog(FormDialog):
    def __init__(self, parent=None, title=None, parent_button_name=None):
        """
        Constructs an advanced form dialog that adds widgets on its parent.

        To add widgets to the parent, call `displayWidgetValueOnParent` after creating an instance
        of the class. The advanced form dialog has a default button in its vertical layout, which
        is located between the form layout and the buttons 'ok' and 'cancel'. This button sets the
        widgets to their default values.

        Parameters
        ----------
        parent : UIFormWidget or None, optional
            The parent widget of the advanced form dialog.
            If None, the dialog is created without a parent.
        title : str or None, optional
            The title of the advanced form dialog.
        parent_button_name : str or None, optional
            The name of the button opening the advanced form dialog in the parent.
            If passed, the extra widgets will be added under this button in the parent,
            otherwise they are added at the end.
        """
        self.dialog_parent = parent
        self.display_on_parent = []
        if parent_button_name is None:
            self.parent_button_row = -1
        elif parent is None:
            raise ValueError(
                'The parent is None. Set the parent if you want to set the parent button name.')
        else:
            self.parent_button_row = self.dialog_parent.getWidgetRow(parent_button_name)

        FormDialog.__init__(self, parent, title)

        # add default button to vertical layout
        self.default_button = QtWidgets.QPushButton("Set default values")
        self.insertWidgetToVerticalLayout(1, self.default_button)
        self.default_button.clicked.connect(lambda: self._setDefaultValues())

    def _onOk(self):
        """
        Called when the "Ok" button is clicked in the advanced dialog.

        Calls the super-class method `_onOk`, sets the default widgets
        to visible, and adds/updates/removes widgets from the
        parent depending on the the widgets in the advanced dialog.
        """
        super()._onOk()
        self.formWidget.setDefaultWidgetStatesVisibleTrue()
        if self.display_on_parent:
            if self.getSavedWidgetStates() == self.getDefaultWidgetStates():
                self._removeWidgetsFromParent()
            else:
                self._addOrUpdateWidgetsInParent()

    def _addOrUpdateWidgetsInParent(self):
        """
        Adds or updates widgets in the parent form.

        This method iterates over the display_on_parent and adds the widgets to the parent
        form. If a widget already exists in the parent form, it is updated with the most current
        value set in the advanced dialog.
        """
        for index, name in enumerate(self.display_on_parent, start=1):
            widget_row = self.parent_button_row + index if self.parent_button_row != -1 else -1
            value = self.getSavedWidgetStates()[f'{name}_field']['value']
            qwidget = self.getWidget(name, 'field')
            if isinstance(qwidget, QtWidgets.QComboBox):
                value = qwidget.itemText(value)
            if f'{name}_field' not in self.dialog_parent.getWidgets():
                label = str(self.getSavedWidgetStates()[f'{name}_label']['value'])
                self.dialog_parent.insertWidget(widget_row, name, QtWidgets.QLabel(str(value)),
                                                label)
            else:
                self.dialog_parent.getWidget(name, 'field').setText(str(value))

    def _removeWidgetsFromParent(self):
        """
        Removes widgets from the parent form.

        This method iterates over the display_on_parent and removes
        the widgets from the parent.
        """
        for name in self.display_on_parent:
            if f'{name}_field' in self.dialog_parent.getWidgets():
                self.dialog_parent.removeWidget(name)

    def _setDefaultValues(self):
        """
        Sets the widgets in the advanced dialog to their default states.

        Makes the default widget states visible, as often the default states are saved while the
        widgets are not visible. Applies the widget states to the widgets in the form.
        """
        self.formWidget.setDefaultWidgetStatesVisibleTrue()
        self.applyWidgetStates(self.formWidget.default_widget_states)

    def displayWidgetValueOnParent(self, name):
        """
        Adds `name` in a list. The order in which names are added to this
        list reflects the order in which the widgets are added to the parent.
        Raises an error if the parent of the advanced dialog is `None`.

        Parameters
        ----------
        name : str
            The name of the widget in the advanced dialog to be displayed in the parent.
        """
        if self.dialog_parent is None:
            raise KeyError('''The advanced-dialog parent is None.
                           Set the parent if you want to add widgets to it.''')
        self.display_on_parent.append(name)



import os
import unittest
from pathlib import Path
from unittest.mock import patch

from PySide2.QtWidgets import QFileDialog

from eqt.ui.SessionDialogs import (
    AppSettingsDialog,
    ErrorDialog,
    LoadSessionDialog,
    SaveSessionDialog,
    SessionDirectorySelectionDialog,
    WarningDialog,
)

from . import skip_ci


@skip_ci
class TestWarningDialog(unittest.TestCase):
    def test_init(self):
        wd = WarningDialog()
        assert wd is not None

    def test_init_with_params(self):
        message = "This is a test message"
        window_title = "Test Warning Dialog Title"
        detailed_text = "This is a test detailed text"

        wd = WarningDialog(message=message, window_title=window_title, detailed_text=detailed_text)
        self.assertEqual(wd.detailedText(), detailed_text)
        self.assertEqual(wd.text(), message)
        self.assertEqual(wd.windowTitle(), window_title)


@skip_ci
class TestErrorDialog(unittest.TestCase):
    def test_init(self):
        ed = ErrorDialog()
        assert ed is not None

    def test_init_with_params(self):
        message = "This is a test message"
        window_title = "Test Error Dialog Title"
        detailed_text = "This is a test detailed text"

        ed = ErrorDialog(message=message, window_title=window_title, detailed_text=detailed_text)
        self.assertEqual(ed.detailedText(), detailed_text)
        self.assertEqual(ed.text(), message)
        self.assertEqual(ed.windowTitle(), window_title)


@skip_ci
class TestSaveSessionDialog(unittest.TestCase):
    def test_init(self):
        ssd = SaveSessionDialog()
        assert ssd is not None

    def test_init_with_title_param(self):
        title = "Test Save Session Dialog Title"
        ssd = SaveSessionDialog(title=title)
        self.assertEqual(ssd.windowTitle(), title)


@skip_ci
class TestSessionDirectorySelectionDialog(unittest.TestCase):
    def test_init(self):
        sdsd = SessionDirectorySelectionDialog()
        assert sdsd is not None

    def test_select_session_directory_label_when_app_name_not_set(self):
        sdsd = SessionDirectorySelectionDialog()

        self.assertEqual(
            sdsd.getWidget("select_session_directory").text(),
            "Select a session directory to save and retrieve all Sessions:")

    def test_select_session_directory_label_when_app_name_set(self):
        sdsd = SessionDirectorySelectionDialog(app_name="Test App")

        self.assertEqual(
            sdsd.getWidget("select_session_directory").text(),
            "Select a session directory to save and retrieve all Test App Sessions:")

    @patch("PySide2.QtWidgets.QFileDialog.getExistingDirectory")
    def test_browse_for_dir_button_makes_file_dialog_for_getting_dir(self, mock_dialog_call):
        sdsd = SessionDirectorySelectionDialog()
        sdsd.browse_for_dir()
        mock_dialog_call.assert_called_once()

    @patch("PySide2.QtWidgets.QFileDialog.getExistingDirectory")
    def test_browse_button_calls_browse_for_dir(self, mock_dialog_call):
        sdsd = SessionDirectorySelectionDialog()
        sdsd.browse_for_dir = unittest.mock.Mock()
        QFileDialog.getExistingDirectory = unittest.mock.Mock()
        sdsd.getWidget("selected_dir").click()
        sdsd.browse_for_dir.assert_called_once()

    def test_browse_dialog_updates_session_directory_label(self):
        example_dir = "C:\\Users\\test_user\\Documents\\test_dir"
        sdsd = SessionDirectorySelectionDialog()
        QFileDialog.getExistingDirectory = unittest.mock.Mock()
        QFileDialog.getExistingDirectory.return_value = example_dir
        sdsd.browse_for_dir()
        self.assertEqual(
            sdsd.getWidget("selected_dir", "label").text(), os.path.basename(example_dir))

    def test_browse_dialog_updates_selected_dir_attribute(self):
        example_dir = "C:\\Users\\test_user\\Documents\\test_dir"
        sdsd = SessionDirectorySelectionDialog()
        QFileDialog.getExistingDirectory = unittest.mock.Mock()
        QFileDialog.getExistingDirectory.return_value = example_dir
        sdsd.browse_for_dir()
        self.assertEqual(sdsd.selected_dir, example_dir)


@skip_ci
class TestLoadSessionDialog(unittest.TestCase):
    def test_init(self):
        lsd = LoadSessionDialog()
        assert lsd is not None

    def test_init_with_title_param(self):
        title = "Test Load Session Dialog Title"
        lsd = LoadSessionDialog(title=title)
        self.assertEqual(lsd.windowTitle(), title)

    def test_init_with_location_of_session_files_param(self):
        location_of_session_files = Path("~").expanduser() / "some" / "test_dir"
        lsd = LoadSessionDialog(location_of_session_files=location_of_session_files)
        self.assertEqual(
            lsd.getWidget("sessions_directory").text(),
            f"Currently loading sessions from: {location_of_session_files}")


@skip_ci
class TestAppSettingsDialog(unittest.TestCase):
    def test_init(self):
        asd = AppSettingsDialog()
        assert asd is not None



import os

from PySide2 import QtWidgets
from pytest import skip

from eqt.ui import FormDialog

is_ci = any(os.getenv(var, '0').lower() in ('1', 'true') for var in ('CONDA_BUILD', 'CI'))

if is_ci:

    def skip_ci(_):
        def inner(*_, **__):
            skip("Running in CI (no GUI)")

        return inner
else:

    def skip_ci(fn):
        return fn


@skip_ci
def test_import():
    from dialog_example_2_test import DialogTest, MainUI # NOQA:F401 isort:skip


class MainUI(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        QtWidgets.QMainWindow.__init__(self, parent)

        pb = QtWidgets.QPushButton(self)
        pb.setText("Open Dialog with form layout")
        pb.clicked.connect(lambda: self.openFormDialog())

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(pb)
        widg = QtWidgets.QWidget()
        widg.setLayout(layout)
        self.push_button = pb

        self.setCentralWidget(widg)

        self.show()

    def openFormDialog(self):

        dialog = FormDialog(parent=self, title='Example')
        dialog.Ok.clicked.connect(lambda: self.accepted())
        dialog.Cancel.clicked.connect(lambda: self.rejected())

        # Example on how to add elements to the
        # add input 1 as QLineEdit
        qlabel = QtWidgets.QLabel(dialog.groupBox)
        qlabel.setText("Input 1: ")
        qwidget = QtWidgets.QLineEdit(dialog.groupBox)
        qwidget.setClearButtonEnabled(True)
        # finally add to the form widget
        dialog.addWidget(qwidget, qlabel, 'input1', layout='form')

        # add input 2 as QComboBox
        qlabel = QtWidgets.QLabel(dialog.groupBox)
        qlabel.setText("Input 2: ")
        qwidget = QtWidgets.QComboBox(dialog.groupBox)
        qwidget.addItem("option 1")
        qwidget.addItem("option 2")
        qwidget.setCurrentIndex(0)
        qwidget.setEnabled(True)
        # finally add to the form widget
        dialog.addWidget(qwidget, qlabel, 'input2')

        # store a reference
        self.dialog = dialog

        dialog.exec()

    def accepted(self):
        print("accepted")
        print(self.dialog.widgets['input1_field'].text())
        print(self.dialog.widgets['input2_field'].currentText())

        self.dialog.close()

    def rejected(self):
        print("rejected")
        self.dialog.close()



import os
import zipfile


def zip_directory(directory: str, compress: bool = True):
    """
    Zips a directory, optionally compressing it.

    Parameters
    ----------
    directory
        The directory to be zipped.
    compress
        Whether to compress the directory.
    """
    compress_type = zipfile.ZIP_DEFLATED if compress else zipfile.ZIP_STORED
    with zipfile.ZipFile(f'{directory}.zip', 'a') as zipper:
        for r, _, f in os.walk(directory):
            for _file in f:
                filepath = os.path.join(r, _file)
                arcname = os.path.relpath(filepath, directory)
                zipper.write(filepath, arcname, compress_type=compress_type)



import time
from time import sleep

from PySide2 import QtCore
from PySide2.QtCore import Qt, QThreadPool
from PySide2.QtWidgets import QProgressDialog

from ..threading import Worker


class ProgressTimerDialog(QProgressDialog):
    def __init__(self, process_name, cancelText="Cancel", parent=None, flags=None,
                 cancel_method=None):
        if flags is None:
            flags = Qt.WindowFlags()
        labelText = f"Running {process_name}"
        QProgressDialog.__init__(self, labelText, cancelText, 0, 0, parent, flags)

        self.setWindowTitle(process_name)
        self.setMinimumDuration(0)
        # This means the other windows can't be used while this is open:
        self.setWindowModality(QtCore.Qt.ApplicationModal)
        self.setWindowFlag(QtCore.Qt.WindowCloseButtonHint, False)
        self.setWindowFlag(QtCore.Qt.WindowMaximizeButtonHint, False)
        self.setWindowFlag(QtCore.Qt.WindowMinimizeButtonHint, False)
        if cancel_method is None:
            self.setCancelButton(None)
        else:
            self.canceled.connect(cancel_method)

        self.threadpool = QThreadPool()
        self.process_name = process_name
        self.run_cancelled = False

    def update_progress_bar(self, value):
        self.setLabelText(f"Running {self.process_name} ... {value}s")

    def show(self):
        QProgressDialog.show(self)
        worker = Worker(self.timing_process)
        worker.signals.progress.connect(self.update_progress_bar)
        self.worker = worker
        self.threadpool.start(worker)

    def timing_process(self, **kwargs):
        progress_callback = kwargs.get('progress_callback')
        t0 = time.time()
        while not self.run_cancelled:
            progress_callback.emit(round(time.time() - t0))
            sleep(1)

    def close(self):
        # If we don't cause timing_process to stop
        # running then it continues forever,
        # even after the progress window closes.
        self.run_cancelled = True
        # need to wait for the thread to finish:
        self.threadpool.waitForDone()
        QProgressDialog.close(self)



from PySide2 import QtWidgets
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QGroupBox, QHBoxLayout, QPushButton


class UIMultiStepWidget(object):
    '''
             QWidget or QDockWidget
    +----------------------------------------------------------+
    |        QVBoxLayout                                       |
    |                                                          |
    |                                                          |
    |                                                          |
    |                                                          |
    |                                                          |
    |                                                          |
    |       _______________               _____________        |
    |      | Previous Step |             | Next Step   |       |
    +----------------------------------------------------------+
    '''
    def createWidget(self):
        # Add vertical layout to dock contents
        verticalLayout = QtWidgets.QVBoxLayout(self)
        verticalLayout.setContentsMargins(10, 10, 10, 10)
        verticalLayout.setAlignment(Qt.AlignTop)

        # Add vertical layout to main widget (self)
        self.setLayout(verticalLayout)

        # Add the next and previous buttons:
        horizontalGroupBox = QGroupBox()
        horizontalGroupBox.setContentsMargins(0, 0, 0, 0)
        horizontalLayout = QHBoxLayout()
        horizontalGroupBox.setLayout(horizontalLayout)
        horizontalGroupBox.setFlat(True)
        verticalLayout.addWidget(horizontalGroupBox)

        next_button = QPushButton("Next step")
        prev_button = QPushButton("Previous step")
        prev_button.setEnabled(False)
        prev_button.clicked.connect(lambda: self.updateStep('prev'))
        next_button.clicked.connect(lambda: self.updateStep('next'))
        horizontalLayout.addWidget(prev_button)
        horizontalLayout.addWidget(next_button)
        self.next_button = next_button
        self.prev_button = prev_button

        self.current_step = -1

        self.uiElements = {'verticalLayout': verticalLayout, 'buttonGroupBox': horizontalGroupBox}
        self.widgets = {}

    def updateStep(self, go_to="next"):
        steps = list(self.widgets)
        step_widgets = list(self.widgets.values())
        current_step_index = steps.index(self.current_step)

        if go_to == 'next':
            self.prev_button.setEnabled(True)
            next_index = current_step_index + 1
            self.next_button.setEnabled(next_index != len(steps) - 1)
        elif go_to == 'prev':
            self.next_button.setEnabled(True)
            next_index = current_step_index - 1
            self.prev_button.setEnabled(next_index != 0)

        self.current_step = steps[next_index]

        widgets_to_show = step_widgets.pop(next_index)
        widgets_to_hide = step_widgets
        self.showHideWidgets(widgets_to_hide, show=False)
        self.showHideWidgets(widgets_to_show, show=True)

    def addStepWidget(self, qwidget, name):
        if self.current_step == -1:
            self.current_step = name
        else:
            self.showHideWidgets(qwidget, show=False)
        self.widgets[name] = qwidget
        # insert widget above next and previous buttons
        self.uiElements['verticalLayout'].insertWidget(0, qwidget)

    def addStepWidgets(self, qwidgets, name):
        for qwidget in qwidgets:
            self.addStepWidget(qwidget, name)

    def showHideWidgets(self, widgets, show=True):
        if not isinstance(widgets, (list, tuple)):
            widgets = [widgets]
        for widget in widgets:
            widget.setVisible(show)


class MultiStepWidget(QtWidgets.QWidget, UIMultiStepWidget):
    def __init__(self, parent=None):
        QtWidgets.QWidget.__init__(self, parent)
        self.createWidget()


class MultiStepDockWidget(QtWidgets.QDockWidget):
    def __init__(self, parent=None, title=None):
        QtWidgets.QDockWidget.__init__(self, parent)
        widget = MultiStepWidget(parent)
        self.setWidget(widget)
        if title is not None:
            self.setObjectName(title)

    def addWidget(self, qwidget, qlabel, name):
        self.widget().addWidget(qwidget, qlabel, name)

    def addStepWidget(self, qwidget, name):
        self.widget().addStepWidget(qwidget, name)

    def addStepWidgets(self, qwidgets, name):
        self.widget().addStepWidgets(qwidgets, name)


class UIMultiStepFactory(QtWidgets.QWidget):
    '''creates a widget with a vertical layout, with "next" and "previous"
    buttons at the bottom of the layout.

    you can add a widget to the vertical layout, using addStepWidget.
    The first step added will be shown to begin with, and the other steps
    can be navigated between, using the next and previous buttons.

    The returned dockWidget must be added with
    main_window.addDockWidget(QtCore.Qt.RightDockWidgetArea, dockWidget)
    '''
    @staticmethod
    def getQDockWidget(parent=None):
        return MultiStepDockWidget(parent)

    @staticmethod
    def getQWidget(parent=None):
        return MultiStepWidget(parent)



import sys

from PySide2 import QtWidgets

from eqt.ui import FormDialog


class MainUI(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        QtWidgets.QMainWindow.__init__(self, parent)

        pb = QtWidgets.QPushButton(self)
        pb.setText("Open Dialog with form layout")
        pb.clicked.connect(lambda: self.openFormDialog())

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(pb)
        widg = QtWidgets.QWidget()
        widg.setLayout(layout)

        self.setCentralWidget(widg)

        self.show()

    def openFormDialog(self):
        dialog = FormDialog(parent=self, title='Example')
        dialog.Ok.clicked.connect(lambda: self.accepted())

        # Example on how to add elements to the FormDialog
        # add input 1 as QLineEdit
        qlabel = QtWidgets.QLabel(dialog.groupBox)
        qlabel.setText("Input 1: ")
        qwidget = QtWidgets.QLineEdit(dialog.groupBox)
        qwidget.setClearButtonEnabled(True)
        # finally add to the form widget
        dialog.addSpanningWidget(QtWidgets.QLabel("Input Values: "), 'input_title')
        dialog.addWidget(qwidget, qlabel, 'input1')

        # add input 2 as QComboBox
        qlabel = QtWidgets.QLabel(dialog.groupBox)
        qlabel.setText("Input 2: ")
        qwidget = QtWidgets.QComboBox(dialog.groupBox)
        qwidget.addItem("option 1")
        qwidget.addItem("option 2")
        qwidget.setCurrentIndex(0)
        qwidget.setEnabled(True)
        # finally add to the form widget
        dialog.addWidget(qwidget, qlabel, 'input2')
        dialog.addWidget(QtWidgets.QLabel("Example Vertical Layout Text"), layout="vertical")

        # Example of using 'getWidget':
        dialog.getWidget('input2').setCurrentIndex(1)

        # store a reference
        self.dialog = dialog
        self.dialog.onCancel = self.rejected
        dialog.exec()

    def accepted(self):
        print("accepted")
        print(self.dialog.widgets['input1_field'].text())
        print(self.dialog.widgets['input2_field'].currentText())
        self.dialog.close()

    def rejected(self):
        print("rejected")


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)

    window = MainUI()

    sys.exit(app.exec_())



import re

import qdarkstyle
from PySide2.QtWidgets import QPushButton, QScrollArea, QWidget


class NoBorderScrollArea(QScrollArea):
    def __init__(self, widget, parent=None):
        """Creates an instance of a QScrollArea and sets its border to none.
        Sets its widget to resizable. Due to a bug in `qdarkstyle`, the PushButtons in the
        widget do not inherit the right style. The init applies `qdarkstyle` to all the buttons
        present in the widget when the object is created. Any button added to the widget after the
        init is invoked will not be styled as expected. In this case, the method
        `apply_qdarkstyle_to_buttons` would need to be invoked by the user after the object is
        instanced."""
        super().__init__(parent)
        self.setStyleSheet("QScrollArea { border: none; }")
        self.setWidgetResizable(True)
        self.setWidget(widget)
        self.apply_qdarkstyle_to_buttons(widget)

    def apply_qdarkstyle_to_buttons(self, widget):
        """Applies the qdarkstyle to all the buttons in the widget explicitly.
        This ensures that the style is consistent with the rest of the app."""
        if isinstance(widget, QPushButton):

            button_style = self._extract_qdarkstyle_button_style()
            widget.setStyleSheet(button_style)
        for child in widget.findChildren(QWidget):
            self.apply_qdarkstyle_to_buttons(child)

    def _extract_qdarkstyle_button_style(self):
        """Returns the QPushButton styles from qdarkstyle, including the different
        button styles."""
        style = qdarkstyle.load_stylesheet(qt_api='pyside2')
        pattern = re.compile(r"(QPushButton\s*{[^}]*}|QPushButton\s*:[^}]*{[^}]*})", re.DOTALL)
        matches = pattern.findall(style)
        if matches:
            return ''.join(matches)
        return ""



import sys
import unittest

from PySide2 import QtWidgets
from PySide2.QtCore import Qt
from PySide2.QtTest import QTest

from eqt.ui import FormDialog

from . import skip_ci


class MainUI(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        QtWidgets.QMainWindow.__init__(self, parent)

        pb = QtWidgets.QPushButton(self)
        pb.setText("Open Dialog with form layout")
        pb.clicked.connect(lambda: self.openFormDialog())

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(pb)
        widg = QtWidgets.QWidget()
        widg.setLayout(layout)
        self.push_button = pb

        self.setCentralWidget(widg)

        self.show()

    def openFormDialog(self):
        dialog = FormDialog(parent=self, title='Example')
        dialog.Ok.clicked.connect(lambda: self.accepted())
        dialog.Cancel.clicked.connect(lambda: self.rejected())

        # Example on how to add elements to the
        # add input 1 as QLineEdit
        qlabel = QtWidgets.QLabel(dialog.groupBox)
        qlabel.setText("Input 1: ")
        qwidget = QtWidgets.QLineEdit(dialog.groupBox)
        qwidget.setClearButtonEnabled(True)
        # finally add to the form widget
        dialog.addWidget(qwidget, qlabel, 'input1', layout='form')

        # add input 2 as QComboBox
        qlabel = QtWidgets.QLabel(dialog.groupBox)
        qlabel.setText("Input 2: ")
        qwidget = QtWidgets.QComboBox(dialog.groupBox)
        qwidget.addItem("option 1")
        qwidget.addItem("option 2")
        qwidget.setCurrentIndex(0)
        qwidget.setEnabled(True)
        # finally add to the form widget
        dialog.addWidget(qwidget, qlabel, 'input2')

        # store a reference
        self.dialog = dialog

        dialog.open()

    def accepted(self):
        print("accepted")
        print(self.dialog.widgets['input1_field'].text())
        print(self.dialog.widgets['input2_field'].currentText())

        self.dialog.close()

    def rejected(self):
        print("rejected")
        self.dialog.close()


_instance = None


class DialogTest(unittest.TestCase):
    '''Test the margarita mixer GUI'''
    @skip_ci
    def setUp(self):
        '''Create the GUI'''
        super(DialogTest, self).setUp()

        global _instance
        if _instance is None:
            _instance = QtWidgets.QApplication(sys.argv)

        self.app = _instance
        window = MainUI()
        self.app = _instance
        self.window = window
        # QTest.mouseClick(self.window.push_button, Qt.LeftButton)
        # self.dialog = window.dialog

    @skip_ci
    def tearDown(self):
        del self.app
        super(DialogTest, self).tearDown()

    @skip_ci
    def test_close(self):
        self.window.close()
        self.assertTrue(True)

    @skip_ci
    def test_openclose_dialog(self):
        QTest.mouseClick(self.window.push_button, Qt.LeftButton)
        dialog = self.window.dialog
        print(dialog)
        dialog.close()

    def stest_defaults(self):
        '''Test the GUI in its default state'''
        self.dialog = self.window.dialog
        print("test1")
        self.assertEqual(self.window.dialog.widgets['input1_field'].text(), '')
        print("test2")
        self.assertEqual(self.window.dialog.widgets['input2_field'].currentIndex(), 0)
        print("click")
        print(self.window.dialog.Ok, self.window.dialog.Cancel)
        QTest.mouseClick(self.window.push_button, Qt.LeftButton)



import sys

from PySide2 import QtWidgets
from PySide2.QtWidgets import QPushButton

from eqt.ui.NoBorderScrollArea import NoBorderScrollArea


class MainUI(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        """Any button added after the `NoBorderScrollArea` is instanced will not be styled as
        expected, due to the bug in `qdarkstyle`. The method `apply_qdarkstyle_to_buttons` needs
        to be invoked."""
        QtWidgets.QMainWindow.__init__(self, parent)

        layout = QtWidgets.QVBoxLayout()
        widg = QtWidgets.QWidget()
        widg.setLayout(layout)

        layout.addWidget(QPushButton("Test"))
        layout.addWidget(QPushButton("Test2"))

        self.scroll_area_widget = NoBorderScrollArea(widg)

        layout.addWidget(QPushButton("Test3"))
        self.scroll_area_widget.apply_qdarkstyle_to_buttons(widg)
        layout.addWidget(QPushButton("Test4"))

        self.setCentralWidget(self.scroll_area_widget)

        self.show()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)

    window = MainUI()

    sys.exit(app.exec_())



from PySide2 import QtCore, QtWidgets


class ReOrderableListWidget(QtWidgets.QTableWidget):
    def __init__(self, parent=None):
        super(ReOrderableListWidget, self).__init__(parent)
        self.setColumnCount(1)
        self.verticalHeader().setVisible(False)
        self.horizontalHeader().setVisible(False)
        self.verticalHeader().setDefaultSectionSize(35)
        self.horizontalHeader().setStretchLastSection(True)
        self.resizeColumnsToContents()

    def addItem(self, name):
        # new row
        row = self.rowCount()
        self.insertRow(row)

        form_widget = QtWidgets.QWidget()

        verticalLayout = QtWidgets.QVBoxLayout(form_widget)
        # Add vertical layout to main widget
        form_widget.setLayout(verticalLayout)
        # Add group box
        groupBox = QtWidgets.QGroupBox(form_widget)
        # Add horizontal layout to group box
        layout = QtWidgets.QHBoxLayout(groupBox)
        # Add elements to layout
        verticalLayout.addWidget(groupBox)

        # button in column 0
        button = QtWidgets.QPushButton()
        button.setProperty("name", name + "_delete")
        pixmapi = QtWidgets.QStyle.SP_DialogCancelButton
        icon = self.style().standardIcon(pixmapi)
        button.setIcon(icon)
        button.clicked.connect(self.onClickDelete)
        button.setMaximumSize(button.minimumSizeHint())
        verticalLayout.setContentsMargins(1, -0.9, 1, -0.9)
        layout.setContentsMargins(2, 1, 2, 1)

        layout.addWidget(QtWidgets.QLabel(name))

        # Up and down buttons:
        button_up = QtWidgets.QPushButton()
        button_up.setProperty("name", name + "_up")
        pixmapi = QtWidgets.QStyle.SP_ArrowUp
        icon = self.style().standardIcon(pixmapi)
        button_up.setIcon(icon)
        button_up.clicked.connect(self.onClickUp)
        button_up.setMaximumSize(button_up.minimumSizeHint())

        button_down = QtWidgets.QPushButton()
        button_down.setProperty("name", name + "_down")
        pixmapi = QtWidgets.QStyle.SP_ArrowDown
        icon = self.style().standardIcon(pixmapi)
        button_down.setIcon(icon)
        button_down.clicked.connect(self.onClickDown)
        button_down.setMaximumSize(button_down.minimumSizeHint())

        layout.addWidget(button_up)
        layout.addWidget(button_down)
        layout.addWidget(button)

        self.setCellWidget(row, 0, form_widget)

        item = QtWidgets.QTableWidgetItem(name)
        # text in column 1
        self.setItem(row, 0, item)

    def addItems(self, names):
        for name in names:
            self.addItem(name)

    def onClickDelete(self):
        # find the item with the same name to get the row
        text = self.sender().property("name")
        item = self.findItems(text.replace("_delete", ""), QtCore.Qt.MatchExactly)[0]
        self.removeRow(item.row())

    def onClickUp(self):
        # find the item with the same name to get the row
        text = self.sender().property("name")
        item = self.findItems(text.strip("_up"), QtCore.Qt.MatchExactly)[0]
        current_row = item.row()
        if item.row() <= 0 or item.row() > (self.rowCount() - 1):
            pass
        else:
            new_row = current_row - 1
            item_to_move_up = self.takeItem(current_row, 0)
            widget_to_move_up = self.cellWidget(current_row, 0)
            self.insertRow(new_row)
            self.setItem(new_row, 0, item_to_move_up)
            self.setCellWidget(new_row, 0, widget_to_move_up)
            self.removeRow(current_row + 1)

    def onClickDown(self):
        # find the item with the same name to get the row
        text = self.sender().property("name")
        item = self.findItems(text.replace("_down", ""), QtCore.Qt.MatchExactly)[0]
        current_row = item.row()
        if item.row() < 0 or item.row() >= self.rowCount() - 1:
            pass
        else:
            new_row = current_row + 2
            item_to_move_up = self.takeItem(current_row, 0)
            widget_to_move_up = self.cellWidget(current_row, 0)
            self.insertRow(new_row)
            self.setItem(new_row, 0, item_to_move_up)
            self.setCellWidget(new_row, 0, widget_to_move_up)
            self.removeRow(current_row)

    def getItemNames(self):
        text_list = []
        for row in range(self.rowCount()):
            item = self.item(row, 0)
            text_list.append(item.text())
        return text_list


class ReOrderableListDockWidget(QtWidgets.QDockWidget):
    def __init__(self, parent=None):
        super(ReOrderableListDockWidget, self).__init__(parent)
        # Make a vertical layout
        self.setWidget(ReOrderableListWidget())



import sys

from PySide2 import QtWidgets

from eqt.ui import FormDialog, UIFormWidget


class MainUI(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        QtWidgets.QMainWindow.__init__(self, parent)

        # create a FormDockWidget
        dock = UIFormWidget.FormDockWidget(parent=self, title='Example remove widget')
        self.addWidgetsToExampleForm(dock)

        # add a button to dock to remove user selected widget
        buttonuser = QtWidgets.QPushButton(dock)
        buttonuser.setText("Remove user selected widget")
        dock.addSpanningWidget(buttonuser, 'Button Remove User')
        buttonuser.clicked.connect(lambda: self.remove(dock, buttonuser))

        # create button for Form Dialog
        pb = QtWidgets.QPushButton(self)
        pb.setText("Open Form Dialog")
        pb.clicked.connect(lambda: self.openFormDialog())

        # create window layout
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(pb)
        layout.addWidget(dock)
        widg = QtWidgets.QWidget()
        widg.setLayout(layout)
        self.setCentralWidget(widg)

        # print dictionary of all widgets in dock
        print("\nDictionary of widgets in the Form Dock Widget:\n" + str(dock.getWidgets()))

        self.show()

    def openFormDialog(self):
        dialog = FormDialog(parent=self, title='Example remove widget')
        dialog.Ok.clicked.connect(lambda: self.remove(dialog, dialog.Ok))
        self.addWidgetsToExampleForm(dialog)
        buttonremovevertical = QtWidgets.QPushButton()
        buttonremovevertical.setText("Remove widget in vertical layout")
        dialog.addSpanningWidget(buttonremovevertical, 'Button remove vertical')
        vertical_button = QtWidgets.QPushButton("Widget in vertical layout")
        buttonremovevertical.clicked.connect(lambda: self.remove_vertical(vertical_button))
        dialog.insertWidgetToVerticalLayout(1, vertical_button)
        dialog.addSpanningWidget(
            QtWidgets.QLabel(
                "Press `Ok` to remove the user selected widget and `Cancel` to close the dialog:"),
            'ok_cancel_instructions')

        # store a reference
        self.dialog = dialog
        self.dialog.onCancel = self.rejected
        # redefine `onOk`` so it does not close the dialog.
        self.dialog._onOk = self._onOkRedefined

        # print dictionary of all widgets in dialog
        print("\nDictionary of widgets in Form Dialog:\n" + str(self.dialog.getWidgets()))

        dialog.open()

    def _onOkRedefined(self):
        '''Saves the widget states.'''
        self.dialog.saveAllWidgetStates()

    def addWidgetsToExampleForm(self, form):

        # add widget 1 as QLineEdit
        qlabel = QtWidgets.QLabel(form)
        qlabel.setText("Widget 1: ")
        qwidget = QtWidgets.QLineEdit(form)
        qwidget.setClearButtonEnabled(True)
        form.addWidget(qwidget, qlabel, 'Widget 1')

        # add widget 2 as QLineEdit
        qlabel = QtWidgets.QLabel(form)
        qlabel.setText("Widget 2: ")
        qwidget = QtWidgets.QLineEdit(form)
        qwidget.setClearButtonEnabled(True)
        form.addWidget(qwidget, qlabel, 'Widget 2')

        # add widget 3 as QLineEdit
        qlabel = QtWidgets.QLabel(form)
        qlabel.setText("Widget 3: ")
        qwidget = QtWidgets.QLineEdit(form)
        qwidget.setClearButtonEnabled(True)
        form.addWidget(qwidget, qlabel, 'Widget 3')

        # add input as QComboBox
        form.addSpanningWidget(QtWidgets.QLabel("Pick the widget you want to remove:"),
                               'input_title')
        qlabel = QtWidgets.QLabel(form)
        qlabel.setText("User input: ")
        qwidget = QtWidgets.QComboBox(form)
        qwidget.addItem("Widget 2")
        qwidget.addItem("Widget 3")
        qwidget.setCurrentIndex(0)
        qwidget.setEnabled(True)
        form.addWidget(qwidget, qlabel, 'userinput')

        # add a button to remove widget 1
        button1 = QtWidgets.QPushButton(form)
        button1.setText("Remove widget 1")
        form.addSpanningWidget(button1, 'Button Remove w1')
        button1.clicked.connect(lambda: self.remove(form, button1, 'Widget 1'))

        # add a button to remove spanning widget
        buttonspanning = QtWidgets.QPushButton(form)
        buttonspanning.setText("Remove spanning widget")
        form.addSpanningWidget(buttonspanning, 'Button Remove Spanning')
        buttonspanning.clicked.connect(lambda: self.remove(form, buttonspanning, 'input_title'))

    def remove_vertical(self, button):
        widget = self.dialog.removeWidgetFromVerticalLayout(button)
        print(f'\nRemoved widget in the vertical layout is {widget}.')
        self.dialog.getWidget('Button remove vertical').setEnabled(False)

    def rejected(self):
        print("\nDialog closed.")

    def remove(self, form, button, userselection=False):
        if userselection is False:
            userselection = form.getWidget('userinput').currentText()
            form.getWidget('userinput').removeItem(form.getWidget('userinput').currentIndex())

        widget = form.removeWidget(userselection)
        print(f'\nRemove {userselection} returning {widget}.')
        if form.getWidget('userinput').currentIndex() == -1:
            button.setEnabled(False)
        if button == form.getWidget('Button Remove w1') or button == form.getWidget(
                'Button Remove Spanning'):
            button.setEnabled(False)

        print("\nDictionary of widgets after deletion of " + userselection + ":\n" +
              str(form.getWidgets()))


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)

    window = MainUI()

    sys.exit(app.exec_())



import shutil

from pytest import fixture, mark

from eqt.io import zip_directory


@fixture
def test_file(tmp_path):
    """Create a zip file, which contains a session.json file"""
    subfile = tmp_path / "Test Subfolder" / "test_file.txt"
    subfolder = subfile.parent
    subfolder.mkdir(parents=True, exist_ok=True)
    subfile.write_text("test")
    return subfile


@mark.parametrize('compress', (True, False))
def test_zip_directory_compress_False(test_file, compress):
    folder = test_file.parent.parent
    zip_directory(folder, compress=compress)
    zipname = folder.with_suffix(".zip")
    assert zipname.is_file()
    shutil.unpack_archive(zipname, folder / "extracted")
    assert (folder / "extracted" / test_file.parts[-2] / test_file.parts[-1]).is_file()



import sys

import utilitiesForExamples
from PySide2 import QtWidgets

from eqt.ui.FormDialog import AdvancedFormDialog
from eqt.ui.UIFormWidget import FormWidget


def run_example():
    parent = FormWidget(parent=None)
    # open advanced dialog button
    pb = QtWidgets.QPushButton(parent)
    pb.setText("Open Advanced Dialog")
    pb.clicked.connect(lambda: advanced_dialog.exec())
    parent.addSpanningWidget(pb, 'button_advanced')
    # extra button
    extra_button = QtWidgets.QPushButton(parent)
    extra_button.setText("Extra button")
    parent.addSpanningWidget(extra_button, 'extra_button')
    # create dialog
    advanced_dialog = AdvancedFormDialog(parent=parent, title='Example',
                                         parent_button_name='button_advanced')
    # all widgets
    utilitiesForExamples.addWidgetsToExample(advanced_dialog)
    for name in utilitiesForExamples.list_all_widgets():
        advanced_dialog.displayWidgetValueOnParent(name)
    # show
    parent.show()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    run_example()
    sys.exit(app.exec_())



import unittest
from unittest import mock

from eqt.ui.MainWindowWithSessionManagement import MainWindowWithSessionManagement
from eqt.ui.SessionDialogs import LoadSessionDialog

from . import skip_ci

# Unit tests for the MainWindowWithSessionManagement class
# methods which create dialogs


@skip_ci
class TestMainWindowWithSessionManagementCreateLoadSessionDialog(unittest.TestCase):
    '''
    Tests the createLoadSessionDialog method of the MainWindowWithSessionManagement class

    This method is responsible for creating the load session dialog
    and populating it with the available sessions
    '''
    def setUp(self):
        self.title = "title"
        self.app_name = "app_name"
        self.smw = MainWindowWithSessionManagement(self.title, self.app_name)
        self.smw.sessions_directory = mock.MagicMock()
        self.zip_folders = ["zip_folder_1", "zip_folder_2"]

    def test_createLoadSessionDialog_populates_session_dropdown(self):
        dialog = self.smw.createLoadSessionDialog(self.zip_folders)
        assert dialog is not None
        assert isinstance(dialog, LoadSessionDialog)
        select_session_combo = dialog.getWidget('select_session')
        assert select_session_combo is not None
        items_in_combo = [
            select_session_combo.itemText(i) for i in range(select_session_combo.count())]
        assert items_in_combo == self.zip_folders

    def test_createLoadSessionDialog_connections(self):
        dialog = self.smw.createLoadSessionDialog(self.zip_folders)

        self.smw.loadSessionLoad = mock.MagicMock()
        self.smw.selectLoadSessionsDirectorySelectedInSessionSelector = mock.MagicMock()
        self.smw.loadSessionNew = mock.MagicMock()

        dialog = self.smw.createLoadSessionDialog(self.zip_folders)

        dialog.Ok.click()
        self.smw.loadSessionLoad.assert_called_once()

        dialog.Select.click()
        self.smw.selectLoadSessionsDirectorySelectedInSessionSelector.assert_called_with(dialog)

        dialog.Cancel.click()
        self.smw.loadSessionNew.assert_called_once()



import sys
from time import sleep

from PySide2 import QtCore, QtWidgets
from PySide2.QtCore import QThreadPool

from eqt.threading import Worker
from eqt.ui import ProgressTimerDialog
from eqt.ui.SessionDialogs import ErrorDialog


class MainUI(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        QtWidgets.QMainWindow.__init__(self, parent)

        pb = QtWidgets.QPushButton(self)
        pb.setText("Start Process")
        pb.clicked.connect(lambda: self.startProcess())

        pb2 = QtWidgets.QPushButton(self)
        pb2.setText("Start Process 2")
        pb2.clicked.connect(lambda: self.startProcessWithError())

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(pb)
        layout.addWidget(pb2)
        widg = QtWidgets.QWidget()
        widg.setLayout(layout)

        self.setCentralWidget(widg)

        self.threadpool = QThreadPool()

        self.show()

    def startProcess(self):
        process_name = "My example process"
        self.create_timer_progress_window(process_name)
        worker = Worker(self.example_process)
        worker.signals.finished.connect(self.progress_windows[process_name].close)
        self.threadpool.start(worker)

    def startProcessWithError(self):
        process_name = "My example process, with an error"
        self.create_timer_progress_window(process_name)
        worker = Worker(self.example_process_with_error)

        worker.signals.finished.connect(self.progress_windows[process_name].close)
        worker.signals.error.connect(self.process_error_dialog)
        self.threadpool.start(worker)

    def example_process(self, **kwargs):
        sleep(10)

    def example_process_with_error(self, **kwargs):
        sleep(3)
        raise Exception("Error!!")

    def process_error_dialog(self, error, **kwargs):
        dialog = ErrorDialog(self, "Error", str(error[1]), str(error[2]))
        dialog.open()

    def create_timer_progress_window(self, process_name):
        if not hasattr(self, 'progress_windows'):
            self.progress_windows = {}

        progress_window = ProgressTimerDialog(process_name, parent=self, flags=QtCore.Qt.Window)

        self.progress_windows[process_name] = progress_window
        progress_window.show()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)

    window = MainUI()

    sys.exit(app.exec_())



from warnings import warn

from PySide2 import QtWidgets

from .UISliderWidget import UISliderWidget


class UIFormWidget:
    '''
             QWidget or QDockWidget
    +----------------------------------------------------------+
    |        QVBoxLayout                                       |
    |   +---------------------------------------------------+  |
    |   |    QGroupBox                                      |  |
    |   |                                                   |  |
    |   |    +------------------------------------------+   |  |
    |   |    |   QFormLayout                            |   |  |
    |   |    |                                          |   |  |
    |   |    |                                          |   |  |
    |   |    +------------------------------------------+   |  |
    |   |                                                   |  |
    |   +---------------------------------------------------+  |
    |                                                          |
    +----------------------------------------------------------+
    '''
    def createForm(self):
        # Add vertical layout to dock contents
        verticalLayout = QtWidgets.QVBoxLayout(self)
        verticalLayout.setContentsMargins(10, 10, 10, 10)

        # Add vertical layout to main widget (self)
        # verticalLayout.addWidget(self)
        self.setLayout(verticalLayout)

        # Add group box
        groupBox = QtWidgets.QGroupBox(self)

        # Add form layout to group box
        groupBoxFormLayout = QtWidgets.QFormLayout(groupBox)

        # Add elements to layout
        verticalLayout.addWidget(groupBox)

        self.uiElements = {
            'verticalLayout': verticalLayout, 'groupBox': groupBox,
            'groupBoxFormLayout': groupBoxFormLayout}
        self.widgets = {}
        self.widget_states = {}
        self.default_widget_states = {}

    @property
    def num_widgets(self):
        return self.uiElements['groupBoxFormLayout'].rowCount()

    @property
    def groupBox(self):
        return self.uiElements['groupBox']

    def insertWidget(self, row, name, qwidget, qlabel=None):
        '''
        Inserts a labelled widget, or a spanning widget, to the form layout.
        The position in the form is specified by row. If row is out of bounds, the widget
        is added at the end of the form. An error is raised if `name` is already in use
        by another widget in the form. The entries associated with the widget are added
        to the widget dictionary and the default-widget-states
        dictionary.

        Parameters
        ----------
        row : int
            The position in the form where the widget is added.
        name : str
            The string associated to the qwidget and qlabel.
        qwidget : QWidget
            The widget to be added on the right hand side of the form or as a spanning widget.
        qlabel : qlabel widget or str
            The qlabel widget, or a str from which a qlabel widget is created, to be added
            on the left hand side of the form. If qlabel is `None` the widget spans the full
            width of the form.
        '''
        if f'{name}_field' in self.widgets:
            raise KeyError(f"Widget name ({name}) already defined. Choose another name.")
        formLayout = self.uiElements['groupBoxFormLayout']
        if formLayout.indexOf(qwidget) != -1:
            raise KeyError(f"The widget {qwidget} is already in use. Create another QWidget.")

        if qlabel is not None:
            if isinstance(qlabel, str):
                txt = qlabel
                qlabel = QtWidgets.QLabel(self)
                qlabel.setText(txt)
            else:
                if formLayout.indexOf(qlabel) != -1:
                    raise KeyError(
                        f"The widget {qlabel} is already in use. Create another QLabel.")
            formLayout.insertRow(row, qlabel, qwidget)
            self.widgets[f'{name}_label'] = qlabel
            self.default_widget_states[f'{name}_label'] = self.getWidgetState(name, 'label')
        else:
            formLayout.insertRow(row, qwidget)

        self.widgets[f'{name}_field'] = qwidget
        self.default_widget_states[f'{name}_field'] = self.getWidgetState(name, 'field')

    def _popWidget(self, dictionary, name):
        '''
        Removes the item(s) associated with `name` from a dictionary.

        Parameters
        ----------
        dictionary : dict
            The dictionary from which to remove the items.
        name : str
            The name of the item(s) to be removed.

        Returns
        -------
        qwidget : QWidget
            The removed widget associated with `name`, if it exists in the dictionary.
        qlabel : QLabel, optional
            The removed label associated with `name`, if it exists in the dictionary.

        Raises
        ------
        KeyError
            If no widget associated with the dictionary key `name` or `{name}_field` is found.
        '''
        if name in dictionary:
            qwidget = dictionary.pop(name)
        elif f'{name}_field' in dictionary:
            qwidget = dictionary.pop(f'{name}_field')
        else:
            raise KeyError(
                f'No widget associated with the dictionary key `{name}` or `{name}_field`.')
        if f'{name}_label' in dictionary:
            qlabel = dictionary.pop(f'{name}_label')
            return qwidget, qlabel
        return qwidget

    def addWidget(self, qwidget, qlabel, name):
        '''
        Adds a widget and a label widget at the the end of
        the form layout.

        Parameters
        ----------
        qwidget : QWidget
            The widget to be added on the right hand side of the form.
        qlabel : qlabel widget or str
            The qlabel widget, or a str from which a qlabel widget is created, to be added
            on the left hand side of the form.
        name : str
            The string associated to the qwidget and qlabel.
        '''
        self.insertWidget(-1, name, qwidget, qlabel)

    def addSpanningWidget(self, qwidget, name):
        '''
        Adds a spanning qwidget occupying the full row in the form layout.

        Parameters
        ----------
        qwidget : widget
            The widget to be added on the form.
        name : str
            The string associated to the qwidget.
        '''
        self.insertWidget(-1, name, qwidget)

    def getNumWidgets(self):
        '''
        Returns the number of widgets in the form.
        '''
        return self.num_widgets

    def removeWidget(self, name):
        '''
        Removes the widget with the specified name from the form layout.
        This method delete the qwidget, and qlabel if present, from the widgets dictionary
        and sets their parent to `None`.

        Parameters
        ----------
        name : str
            The name of the widget to be removed.

        Returns
        -------
        tuple or QWidget
            If the widget has a corresponding label, a tuple containing the widget
            and label is returned. Otherwise, only the widget is returned.
        '''
        widget_row = self.getWidgetRow(name)
        self.getWidget(name, 'field').setParent(None)
        if f'{name}_label' in self.widgets:
            self.getWidget(name, 'label').setParent(None)
            qwidget, qlabel = self._popWidget(self.widgets, name)
            self.uiElements['groupBoxFormLayout'].removeRow(widget_row)
            return qwidget, qlabel
        qwidget = self._popWidget(self.widgets, name)
        self.uiElements['groupBoxFormLayout'].removeRow(widget_row)
        return qwidget

    def getWidget(self, name, role='field'):
        '''
        Returns the widget by the name with which it has been added.
        By default it returns the widget that is the field in the form.
        The user can get the label by specifying the role to be label.
        Raises ValueError if the role is not field or label.
        '''
        allowed_roles = 'field', 'label'
        if role in allowed_roles:
            return self.widgets[f'{name}_{role}']
        raise ValueError(f'Unexpected role: expected any of {allowed_roles}, got {role}')

    def getWidgetRow(self, name, role='field'):
        '''
        Returns the widget row in the form layout by the widget name.
        This is the row of the widget in the form layout.
        '''
        return self.uiElements['groupBoxFormLayout'].getWidgetPosition(self.getWidget(name,
                                                                                      role))[0]

    def setWidgetVisible(self, name, visible):
        '''
        Sets the visibility of the widget and associated label with the given name.

        Parameters
        ----------
        name : str
            The name of the widget to set visible/invisible
        visible : bool
            True to set the widget visible, False to hide it
        '''
        allowed_roles = ['field', 'label']
        for role in allowed_roles:
            try:
                self.getWidget(name, role).setVisible(visible)
            except Exception:
                # We may not have a label for the widget
                pass

    def getWidgets(self):
        '''Returns a dictionary of the widgets currently present in the form.'''
        return self.widgets

    def addTitle(self, qlabel, name):
        if isinstance(qlabel, str):
            txt = qlabel
            qlabel = QtWidgets.QLabel(self.uiElements['groupBox'])
            qlabel.setText(txt)
        qlabel.setStyleSheet("font-weight: bold")
        self.insertWidget(-1, name, qlabel)

    def addSeparator(self, name):
        # Adds horizontal separator to the form
        frame = QtWidgets.QFrame()
        frame.setFrameShape(QtWidgets.QFrame.HLine)
        frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.insertWidget(-1, name, frame)

    def setDefaultWidgetStatesVisibleTrue(self):
        '''
        Sets all of the entries 'visible' in the `default_widget_states` dictionary to be `True`.
        '''
        for state in self.default_widget_states.values():
            state['visible'] = True

    def getAllWidgetStates(self):
        '''
        Returns
        -------
        dict
            Format: {'widget_name': {'value': str | bool | int, 'enabled': bool, 'visible': bool,
            'widget_row': int}, ...}.
            e.g. {'widget1': {'value': 1, 'enabled': True, 'visible': True, 'widget_row': 0},
            'widget2': {'value': 2, 'enabled': False, 'visible': False, 'widget_row': 1}}.
        '''
        return {key: self.getWidgetState(widget) for key, widget in self.widgets.items()}

    def getWidgetState(self, widget, role=None):
        '''
        Returns the current state of the widget in the form.

        Parameters
        ----------
        widget : QWidget or str
            The widget or its name (or its name + '_field' or '_label', when role is None) to get
            the state of.
        role : str, optional, default None, values: 'label', 'field', None.
            The role of the widget to apply the state to. This is used only if `widget` is the
            widget name string.

        Returns
        -------
        state : dict
            Format: {'value': str | bool | int, 'enabled': bool, 'visible': bool,
            'widget_row' : int}.
            e.g. {'value': 1, 'enabled': True, 'visible': True, 'widget_row' : 0}.
            This can be used to restore the state of the widget using `setWidgetState()`.
        '''
        if widget is None:
            raise ValueError('The widget or the name of widget must be given.')

        if isinstance(widget, str):
            if role is not None:
                if role not in ['label', 'field']:
                    raise ValueError(f'Role must be either "label", "field" or None. Got {role}.')
                name_role = widget + '_' + role
                name = widget
            else:
                name, role = self._getNameAndRoleFromKey(widget)
                name_role = name + '_' + role

            try:
                widget = self.widgets[name_role]
            except KeyError:
                raise KeyError(f'No widget associated with the dictionary key `{name_role}`.')
        else:
            name, role = self._getNameAndRoleFromWidget(widget)
        widget_state = {}
        if isinstance(widget, QtWidgets.QLabel):
            widget_state['value'] = widget.text()
        elif isinstance(widget, (QtWidgets.QCheckBox, QtWidgets.QPushButton)):
            widget_state['value'] = widget.isChecked()
        elif isinstance(widget, QtWidgets.QComboBox):
            widget_state['value'] = widget.currentIndex()
        elif isinstance(widget, UISliderWidget) or isinstance(widget, QtWidgets.QSlider):
            widget_state['value'] = widget.value()
        elif isinstance(widget, (QtWidgets.QDoubleSpinBox, QtWidgets.QSpinBox)):
            widget_state['value'] = widget.value()
        elif isinstance(widget, QtWidgets.QLineEdit):
            widget_state['value'] = widget.text()
        elif isinstance(widget, QtWidgets.QRadioButton):
            widget_state['value'] = widget.isChecked()
        elif isinstance(widget, (QtWidgets.QTextEdit, QtWidgets.QPlainTextEdit)):
            widget_state['value'] = widget.toPlainText()
        widget_state['enabled'] = widget.isEnabled()
        widget_state['visible'] = widget.isVisible()
        widget_state['widget_row'] = self.getWidgetRow(name, role)
        return widget_state

    def _getNameAndRoleFromKey(self, key):
        '''
        Given a key, returns the name and the role.
        Role can be extracted as the suffix to key or is set as `field` by default.

        Parameters
        -------------
        key : str
            Format: name or name_field or name_label
        '''
        if key.endswith('_field'):
            return key.removesuffix('_field'), 'field'
        elif key.endswith('_label'):
            return key.removesuffix('_label'), 'label'
        return key, 'field'

    def _getNameAndRoleFromWidget(self, widget):
        '''
        Given a widget, finds it in the widget dictionary and returns its name and role.

        Parameters
        -------------
        widget : qwidget
        '''
        for key, value in self.widgets.items():
            if value == widget:
                return self._getNameAndRoleFromKey(key)
        raise KeyError(f'There is no widget {widget} in the form.')

    def applyWidgetState(self, name, state, role=None):
        '''
        Applies the given `state` to the widget associated with `name` and `role`.
        If role is None, the role is assigned to be 'field'.

        Parameters
        ----------
        name : str
            The name of the widget to apply the state to.
        role : str, optional, default None, values: 'label', 'field', None.
            The role of the widget to apply the state to.
        state : dict
            Format: {'value': str | bool | int, 'enabled': bool, 'visible': bool,
            'widget_row' : int}.
            e.g. {'value': 1, 'enabled': True, 'visible': True, 'widget_row' : 0}.
        '''
        if role is not None:
            if role in ('label', 'field'):
                name_role = name + '_' + role
            else:
                raise ValueError(f'Role must be either "label", "field" or None. Got {role}.')
        else:
            name_role = f'{name}_field'

        # retrieve widget
        widget = self.widgets[name_role]
        # apply state
        for key, value in state.items():
            if key == 'enabled':
                widget.setEnabled(value)
            elif key == 'visible':
                widget.setVisible(value)
            elif key == 'value':
                if isinstance(widget, QtWidgets.QLabel):
                    widget.setText(value)
                elif isinstance(widget, QtWidgets.QCheckBox):
                    widget.setChecked(value)
                elif isinstance(widget, QtWidgets.QComboBox):
                    widget.setCurrentIndex(value)
                elif isinstance(widget, (UISliderWidget, QtWidgets.QSlider)):
                    widget.setValue(value)
                elif isinstance(widget, (QtWidgets.QDoubleSpinBox, QtWidgets.QSpinBox)):
                    widget.setValue(value)
                elif isinstance(widget, QtWidgets.QPushButton):
                    widget.setChecked(value)
                elif isinstance(widget, QtWidgets.QLineEdit):
                    widget.setText(value)
                elif isinstance(widget, QtWidgets.QRadioButton):
                    widget.setChecked(value)
                elif isinstance(widget, (QtWidgets.QTextEdit, QtWidgets.QPlainTextEdit)):
                    widget.setPlainText(value)

    def applyWidgetStates(self, states):
        '''
        Applies the given states to the form's widgets. It raises an error if the keys in the
        dicitonary of states and the keys in the dictionary of widgets in the form are not the
        same.

        Parameters
        ----------
        states : dict
            Format: {'name_field': {'value': str | bool | int, 'enabled': bool, 'visible': bool,
            'widget_row' : int}, 'name_label': {'value': str | bool | int, 'enabled': bool,
            'visible': bool, 'widget_row' : int}, ...}.
            e.g. {'widget1': {'value': 1, 'enabled': True, 'visible': True, 'widget_row': 0},
                  'widget2': {'value': 2, 'enabled': False, 'visible': False, 'widget_row': 1}}.
        '''
        if set(self.widgets) != set(states):
            raise KeyError(f"states={set(states)} do not match form widgets ({set(self.widgets)})")
        for key, widget_state in states.items():
            name, role = self._getNameAndRoleFromKey(key)
            self.applyWidgetState(name, widget_state, role)

    def saveAllWidgetStates(self):
        '''
        Saves the state of all widgets currently present in the form.
        To later restore the states, use `restoreAllSavedWidgetStates()`.
        '''
        self.widget_states = self.getAllWidgetStates()

    def getWidgetStates(self):
        '''Deprecated. Use `getSavedWidgetStates`.'''
        warn('The method `getWidgetStates` is deprecated, use `getSavedWidgetStates`.',
             DeprecationWarning, stacklevel=2)
        return self.getSavedWidgetStates()

    def getSavedWidgetStates(self):
        '''Returns the saved widget states.'''
        return self.widget_states

    def getDefaultWidgetStates(self):
        '''Returns the saved default widget states.'''
        return self.default_widget_states

    def restoreAllSavedWidgetStates(self):
        '''
        All widgets in the form are restored to the saved states. There are saved states only if
        `saveAllWidgetStates` was previously invoked. If there are no previously saved states,
        `default_widget_states` are used instead, after being made visible.
        '''
        if not self.widget_states:
            self.setDefaultWidgetStatesVisibleTrue()
            self.applyWidgetStates(self.default_widget_states)
        else:
            self.applyWidgetStates(self.widget_states)


class FormWidget(QtWidgets.QWidget, UIFormWidget):
    def __init__(self, parent=None):
        # dockWidgetContents = QtWidgets.QWidget()

        QtWidgets.QWidget.__init__(self, parent)
        self.createForm()


class FormDockWidget(QtWidgets.QDockWidget):
    def __init__(self, parent=None, title=''):
        if title is None:
            title = ''
        super().__init__(title, parent)
        widget = FormWidget(parent)
        self.setWidget(widget)
        if title is not None:
            self.setObjectName(title)

    def addWidget(self, qwidget, qlabel, name):
        '''
        Adds a widget and a label widget at the the end of
        the form layout.

        Parameters
        ----------
        qwidget : widget
            The widget to be added on the right hand side of the form.
        qlabel : qlabel widget or str
            The qlabel widget, or a str from which a qlabel widget is created, to be added
            on the left hand side of the form.
        name : str
            The string associated to the qwidget and qlabel.
        '''
        self.widget().addWidget(qwidget, qlabel, name)

    def addSpanningWidget(self, qwidget, name):
        '''
        Adds a spanning qwidget occupying the full row in the form layout.

        Parameters
        ----------
        qwidget : widget
            The widget to be added on the form.
        name : str
            The string associated to the qwidget.
        '''
        self.widget().addSpanningWidget(qwidget, name)

    def insertWidget(self, row, name, qwidget, qlabel=None):
        '''
        Inserts a labelled widget, or a spanning widget, to the form layout.
        The position in the form is specified by row. If row is out of bounds, the widget
        is added at the end of the form. An error is raised if `name` is already in use
        by another widget in the form. The entries associated with the widget are added
        to the widget dictionary and the default-widget-states
        dictionary.

        Parameters
        ----------
        row : int
            The position in the form where the widget is added.
        name : str
            The string associated to the qwidget and qlabel.
        qwidget : QWidget
            The widget to be added on the right hand side of the form or as a spanning widget.
        qlabel : qlabel widget or str
            The qlabel widget, or a str from which a qlabel widget is created, to be added
            on the left hand side of the form. If qlabel is `None` the widget spans the full
            width of the form.
        '''
        self.widget().insertWidget(row, name, qwidget, qlabel)

    def removeWidget(self, name):
        '''
        Removes the widget with the specified name from the form layout.
        This method delete the qwidget, and qlabel if present, from the widgets dictionary
        and sets their parent to `None`.

        Parameters
        ----------
        name : str
            The name of the widget to be removed.

        Returns
        -------
        tuple or QWidget
            If the widget has a corresponding label, a tuple containing the widget
            and label is returned. Otherwise, only the widget is returned.
        '''
        return self.widget().removeWidget(name)

    def getNumWidgets(self):
        '''
        Returns the number of widgets in the form.
        '''
        return self.widget().getNumWidgets()

    def getWidget(self, name, role='field'):
        '''
        Returns the widget by the name with which it has been added.
        By default it returns the widget that is the field in the form.
        The user can get the label by specifying the role to be label.
        Raises ValueError if the role is not field or label.
        '''
        return self.widget().getWidget(name, role)

    def getWidgets(self):
        '''Returns a dictionary of the widgets currently present in the form.'''
        return self.widget().getWidgets()

    def getWidgetRow(self, name, role='field'):
        '''
        Returns the widget row in the form layout by the widget name.
        This is the row of the widget in the form layout.
        '''
        return self.widget().getWidgetRow(name, role)

    def setWidgetVisible(self, name, visible):
        '''
        Sets the visibility of the widget and associated label with the given name.

        Parameters
        ----------
        name : str
            The name of the widget to set visible/invisible
        visible : bool
            True to set the widget visible, False to hide it
        '''
        self.widget().setWidgetVisible(name, visible)

    def saveAllWidgetStates(self):
        '''
        Saves the state of all widgets currently present in the form.
        To later restore the states, use `restoreAllSavedWidgetStates()`.
        '''
        self.widget().saveAllWidgetStates()

    def getWidgetStates(self):
        '''Deprecated. Use `getSavedWidgetStates`.'''
        return self.widget().getWidgetStates()

    def getSavedWidgetStates(self):
        '''Returns the saved widget states.'''
        return self.widget().getSavedWidgetStates()

    def getDefaultWidgetStates(self):
        '''Returns the saved default widget states.'''
        return self.widget().getDefaultWidgetStates()

    def restoreAllSavedWidgetStates(self):
        '''
        All widgets in the form are restored to the saved states. There are saved states only if
        `saveAllWidgetStates` was previously invoked. If there are no previously saved states,
        `default_widget_states` are used instead, after being made visible.
        '''
        self.widget().restoreAllSavedWidgetStates()

    def getAllWidgetStates(self):
        '''
        Returns
        -------
        dict
            Format: {'widget_name': {'value': str | bool | int, 'enabled': bool, 'visible': bool,
            'widget_row': int}, ...}.
            e.g. {'widget1': {'value': 1, 'enabled': True, 'visible': True, 'widget_row': 0},
            'widget2': {'value': 2, 'enabled': False, 'visible': False, 'widget_row': 1}}.
        '''
        return self.widget().getAllWidgetStates()

    def getWidgetState(self, widget, role=None):
        '''
        Returns the current state of the widget in the form.

        Parameters
        ----------
        widget : QWidget or str
            The widget or its name (or its name + '_field' or '_label', when role is None) to get
            the state of.
        role : str, optional, default None, values: 'label', 'field', None.
            The role of the widget to apply the state to. This is used only if `widget` is the
            widget name string.

        Returns
        -------
        state: dict
            Format: {'value': str | bool | int, 'enabled': bool, 'visible': bool,
            'widget_row' : int}.
            e.g. {'value': 1, 'enabled': True, 'visible': True, 'widget_row' : 0}.
            This can be used to restore the state of the widget using `setWidgetState()`.
        '''
        return self.widget().getWidgetState(widget, role)

    def applyWidgetState(self, name, state, role=None):
        '''
        Applies the given `state` to the widget associated with `name` and `role`.
        If role is None, the role is assigned to be 'field'.

        Parameters
        ----------
        name : str
            The name of the widget to apply the state to.
        role : str, optional, default None, values: 'label', 'field', None.
            The role of the widget to apply the state to.
        state : dict
            Format: {'value': str | bool | int, 'enabled': bool, 'visible': bool,
            'widget_row' : int}.
            e.g. {'value': 1, 'enabled': True, 'visible': True, 'widget_row' : 0}.
        '''
        return self.widget().applyWidgetState(name, state, role)

    def applyWidgetStates(self, states):
        '''
        Applies the given states to the form's widgets. It raises an error if the keys in the
        dicitonary of states and the keys in the dictionary of widgets in the form are not the
        same.

        Parameters
        ----------
        states : dict
            Format: {'name_field': {'value': str | bool | int, 'enabled': bool, 'visible': bool,
                     'widget_row' : int}, 'name_label': {'value': str | bool | int,
                     'enabled': bool, 'visible': bool, 'widget_row' : int}, ...}.
            e.g. {'widget1': {'value': 1, 'enabled': True, 'visible': True, 'widget_row': 0},
                  'widget2': {'value': 2, 'enabled': False, 'visible': False, 'widget_row': 1}}.
        '''
        return self.widget().applyWidgetStates(states)


class UIFormFactory(QtWidgets.QWidget):
    # def generateUIFormView(QtWidgets.QWidget):
    '''creates a widget with a form layout group to add things to

    basically you can add widget to the returned groupBoxFormLayout and paramsGroupBox
    The returned dockWidget must be added with
    main_window.addDockWidget(QtCore.Qt.RightDockWidgetArea, dockWidget)
    '''
    @staticmethod
    def getQDockWidget(parent=None, title=None):
        return FormDockWidget(parent, title)

    @staticmethod
    def getQWidget(parent=None):
        return FormWidget(parent)



import sys

from PySide2 import QtWidgets

from eqt.ui import UIFormFactory, UIMultiStepFactory


class MainUI(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        QtWidgets.QMainWindow.__init__(self, parent)

        pb = QtWidgets.QPushButton(self)
        pb.setText("Open Dialog with MultiStep widget")
        pb.clicked.connect(lambda: self.openMultiStepDialog())

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(pb)
        widg = QtWidgets.QWidget()
        widg.setLayout(layout)

        self.setCentralWidget(widg)

        self.show()

    def openMultiStepDialog(self):

        dialog = QtWidgets.QDialog(parent=self)
        bb = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok
                                        | QtWidgets.QDialogButtonBox.Cancel)

        bb.button(QtWidgets.QDialogButtonBox.Ok).clicked.connect(lambda: self.accepted())
        bb.button(QtWidgets.QDialogButtonBox.Cancel).clicked.connect(lambda: self.rejected())
        self.buttonBox = bb

        verticalLayout = QtWidgets.QVBoxLayout(self)
        verticalLayout.setContentsMargins(10, 10, 10, 10)

        # Add vertical layout to main widget (self)
        dialog.setLayout(verticalLayout)

        # create a form layout widget
        fw1 = UIFormFactory.getQWidget(parent=self)

        # ## Example on how to add elements to the form

        # add title
        qlabel = QtWidgets.QLabel(fw1.groupBox)
        qlabel.setText("Form Widget 1")
        fw1.addTitle(qlabel, 'title')

        # add input 1 as QLineEdit
        qlabel = QtWidgets.QLabel(fw1.groupBox)
        qlabel.setText("Input 1: ")
        qwidget = QtWidgets.QLineEdit(fw1.groupBox)
        qwidget.setClearButtonEnabled(True)
        # finally add to the form widget
        fw1.addWidget(qwidget, qlabel, 'input1')

        # add separator
        fw1.addSeparator('separator_1_2')

        # add input 2 as QComboBox
        qlabel = "Input 2: "
        qwidget = QtWidgets.QComboBox(fw1.groupBox)
        qwidget.addItem("option 1")
        qwidget.addItem("option 2")
        qwidget.setCurrentIndex(0)
        qwidget.setEnabled(True)
        # finally add to the form widget
        fw1.addWidget(qwidget, qlabel, 'input2')

        # create a 2nd form layout widget
        fw2 = UIFormFactory.getQWidget(parent=self)

        # ## Example on how to add elements to the form

        # add title
        qlabel = QtWidgets.QLabel(fw2.groupBox)
        qlabel.setText("Form Widget 2")
        fw2.addTitle(qlabel, 'title')

        # add input 1 as QLineEdit
        qlabel = QtWidgets.QLabel(fw2.groupBox)
        qlabel.setText("Input 3: ")
        qwidget = QtWidgets.QLineEdit(fw2.groupBox)
        qwidget.setClearButtonEnabled(True)
        # finally add to the form widget
        fw2.addWidget(qwidget, qlabel, 'input3')

        multistep_widget = UIMultiStepFactory.getQWidget(parent=self)
        multistep_widget.addStepWidget(fw1, 'fw1')
        multistep_widget.addStepWidget(fw2, 'fw2')

        verticalLayout.addWidget(multistep_widget)
        verticalLayout.addWidget(bb)

        # store a reference
        self.fw1 = fw1
        self.fw2 = fw2
        self.dialog = dialog

        dialog.exec()

    def accepted(self):
        print("accepted")
        print(self.fw1.widgets['input1_field'].text())
        print(self.fw1.widgets['input2_field'].currentText())
        print(self.fw2.widgets['input3_field'].text())

        self.dialog.close()

    def rejected(self):
        print("rejected")
        self.dialog.close()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)

    window = MainUI()

    sys.exit(app.exec_())



import json
import os
import shutil
from datetime import datetime
from functools import partial

from PySide2.QtGui import QCloseEvent, QKeySequence
from PySide2.QtWidgets import QAction

from ..io import zip_directory
from ..threading import Worker
from .MainWindowWithProgressDialogs import MainWindowWithProgressDialogs
from .SessionDialogs import (
    ErrorDialog,
    LoadSessionDialog,
    SaveSessionDialog,
    SessionDirectorySelectionDialog,
)


class MainWindowWithSessionManagement(MainWindowWithProgressDialogs):
    '''
    A base class for a main window that can save and load sessions.

    This class is meant to be subclassed, and the subclass should implement
    the following methods:
    - getSessionConfig
    - finishLoadConfig

    In a derived class's __init__ method, the __init__ method of this class
    should be called first, before any other initialisation.


    Properties of Note:
    -------------------
    self.settings
        This is a QSettings object that is used to save and load the session
        settings such as the light/dark theme, the session directory, etc.
    self.progress_windows
        This is a dictionary of ProgressTimerDialog objects, where the key is
        the name of the progress window.
    self.sessions_directory
        This is the path to the directory where the session folders are saved.
        This is set by the user using the menu option:
        "Settings > Set Session Directory".
        Inside the folder specified by the user, there will be a
        folder called self.sessions_directory_name, which is where
        the session folders are saved.
        So self.sessions_directory will be:
        <user selected directory>/<self.sessions_directory_name>
    '''
    def __init__(self, title, app_name, settings_name=None, organisation_name=None, **kwargs):

        super(MainWindowWithSessionManagement, self).__init__(title, app_name, settings_name,
                                                              organisation_name, **kwargs)

        self._setupMainWindowWithSessionManagement()

    def _setupMainWindowWithSessionManagement(self):
        '''Setup the session main window.

        This is called by the __init__ method, and should not be called by the
        user.
        '''

        # This is the name of the directory where the session folders are saved
        # This will be within a directory that is picked by the user
        # i.e. the location of the session folders will be:
        # <user selected directory>/<self.sessions_directory_name>
        self.sessions_directory_name = self.app_name.replace(" ", "-") + "-Sessions"
        self.sessions_directory = None

        self.setupSession()

        self.should_really_close = False

    # Create the menu ----------------------------------------------------------

    def createMenu(self):
        '''Create the menu bar, with the following options:

        File
            Save
                This will save the current session
            Save & Exit
                This will save the current session and exit the application
            Exit
                This will exit the application without saving the current session

        Settings
            App Settings
                This will open a dialog to change the application settings, such as
                the light/dark theme
            Session Directory
                This will open a dialog to select the path where the session folders
                are saved

        Sets
        ----
        self.menu_bar: QMenuBar
            The menu bar
        self.menus: dict
            A dictionary of the menu names and the corresponding QMenu objects

        '''

        menu_bar = self.menuBar()

        # If we add the menus this way, then the menuBar takes ownership of the menus:
        file_menu = menu_bar.addMenu("File")

        save_action = QAction("Save", self)
        save_action.triggered.connect(partial(self.createSaveWindow, event="SaveEvent"))
        file_menu.addAction(save_action)

        save_exit_action = QAction("Save + Exit", self)
        save_exit_action.triggered.connect(self.close)
        file_menu.addAction(save_exit_action)

        exit_action = QAction("Exit", self)
        exit_action.setShortcut(QKeySequence.Quit)
        exit_action.triggered.connect(self.close)
        file_menu.addAction(exit_action)

        settings_menu = menu_bar.addMenu("Settings")

        app_settings_action = QAction("App Settings", self)
        app_settings_action.triggered.connect(self.createAppSettingsDialog)
        settings_menu.addAction(app_settings_action)

        session_folder_action = QAction("Set Session Directory", self)
        session_folder_action.triggered.connect(
            lambda: self.createSessionsDirectorySelectionDialog(new_session=False))
        settings_menu.addAction(session_folder_action)

        menus = {"File": file_menu, "Settings": settings_menu}

        self.menu_bar = menu_bar
        self.menus = menus

    # Loading Sessions ---------------------------------------------------------

    def setupSession(self):
        '''
        Loads the appropriate dialogs to allow users to make selections about loading sessions:

        If this is the first time they open the application, they will be asked to select
        a directory to store the sessions in. This directory will be stored in the settings,
        and will be used as the default directory for future sessions.
        Then, if any sessions exist already in that directory, they will be asked if they want
        to load one of those sessions, if not, a new session will be created.

        If this is not the first time they open the application, they will be asked to select
        a session to load, if any sessions are present in the directory saved in the settings.
        If they select a session, it will be loaded. If they select "New Session",
        a new session will be created.
        '''

        if self.settings.value('sessions_folder') is None:
            # get user to select directory, then within that directory we create a folder called
            # session_folder_name, which will then contain all of the session folders.
            self.createSessionsDirectorySelectionDialog(new_session=True)
        else:
            session_folder_name = self.settings.value('sessions_folder')
            if not os.path.exists(session_folder_name):
                self.createSessionsDirectorySelectionDialog(new_session=True)
            else:
                self.sessions_directory = session_folder_name
                self.createSessionSelector()

    def createSessionSelector(self):
        ''''
        Create a LoadSessionDialog, populated with the names of the sessions
        saved in the current working directory.

        If no sessions exist, create a new session
        '''
        zip_folders = []
        for _, _, f in os.walk(self.sessions_directory):
            for _file in f:
                if '.zip' in _file:
                    array = _file.split("_")
                    if (len(array) > 1):
                        date = _file.split("_")[-1]
                        name = _file.replace("_" + date, "") + " " + date.strip(".zip")
                        zip_folders.append(name)

        if len(zip_folders) == 0:
            self.loadSessionNew()
        else:
            self.SessionSelectionWindow = self.createLoadSessionDialog(zip_folders)

    def createLoadSessionDialog(self, zip_folders):
        '''
        Create a LoadSessionDialog, populated with the names of the sessions
        saved in the current working directory.
        '''
        dialog = LoadSessionDialog(parent=self, location_of_session_files=self.sessions_directory)
        dialog.widgets['select_session_field'].addItems(zip_folders)
        dialog.Ok.clicked.connect(self.loadSessionLoad)
        dialog.Select.clicked.connect(
            lambda: self.selectLoadSessionsDirectorySelectedInSessionSelector(dialog))
        dialog.Cancel.clicked.connect(self.loadSessionNew)
        dialog.open()

        return dialog

    def selectLoadSessionsDirectorySelectedInSessionSelector(self, load_session_dialog):
        '''
        Called when the user selects the "Select Directory for Loading Sessions" button
        in the LoadSessionDialog. It will close the LoadSessionDialog and
        open a SessionDirectorySelectionDialog to allow the user to
        select a directory in which the list of sessions will be loaded from.
        '''
        load_session_dialog.close()
        self.createSessionsDirectorySelectionDialog(new_session=True)

    def createSessionsDirectorySelectionDialog(self, new_session=False):
        '''
        Create a SessionDirectorySelectionDialog, which allows the user to select
        a directory in which the sessions folder will be created.

        Opens the dialog, and connects the Ok button to the appropriate method

        Parameters
        ----------
        new_session : bool
            Whether or not this method is called straight after the user has opened the application
        '''
        session_folder_selection_dialog = SessionDirectorySelectionDialog(
            parent=self, app_name=self.app_name)

        if new_session:
            # Move on to creating directory and loading session names
            session_folder_selection_dialog.Ok.clicked.connect(
                lambda: self.onSessionDirectorySelectionDialogOkNewSession(
                    session_folder_selection_dialog))
        else:
            # Splitting the path and the folder name,
            # means that we only display the folder name that the user has selected
            # and not the full path including the self.sessions_directory_name:
            session_folder_selection_dialog.Ok.clicked.connect(
                lambda: self.onSessionDirectorySelectionDialogOkInSession(
                    session_folder_selection_dialog))
            session_folder_selection_dialog.Cancel.setEnabled(True)
            session_folder_selection_dialog.Cancel.clicked.connect(
                session_folder_selection_dialog.close)

        if self.sessions_directory is not None:
            session_folder_selection_dialog.getWidget('selected_dir', 'label').setText(
                os.path.split(self.sessions_directory)[0])
            session_folder_selection_dialog.selected_dir = os.path.split(
                self.sessions_directory)[0]

        session_folder_selection_dialog.open()

        return session_folder_selection_dialog

    def onSessionDirectorySelectionDialogOkNewSession(self, session_folder_selection_dialog):
        '''
        Runs if the user selects Ok on the `SessionDirectorySelectionDialog`
        when creating a new session
        '''
        if session_folder_selection_dialog.selected_dir is not None:
            self.createSessionsDirectoryAndMakeSessionSelector(
                session_folder_selection_dialog.selected_dir)
            session_folder_selection_dialog.close()

    def onSessionDirectorySelectionDialogOkInSession(self, session_folder_selection_dialog):
        '''
        Runs if the user selects Ok on the SessionDirectorySelectionDialog when in a session
        '''
        if session_folder_selection_dialog.selected_dir is not None:
            self.createSessionsDirectory(session_folder_selection_dialog.selected_dir)
        session_folder_selection_dialog.close()

    def createSessionsDirectoryAndMakeSessionSelector(self, user_chosen_directory):
        '''
        Creates the sessions directory, and then calls the method to create the session selector
        '''
        self.createSessionsDirectory(user_chosen_directory)
        self.createSessionSelector()

    def createSessionsDirectory(self, user_selected_directory):
        '''
        If the directory selected by the user does not exist, create it,
        and then create a folder called sessions_directory_name within it.
        Move into the sessions_directory_name folder,
        and store the path to that folder in the settings.
        Also stores the path to self.sessions_directory_name
        '''
        # If the user has selected the sessions_directory_name folder,
        # we want to move up a directory before looking for the
        # sessions_directory_name folder within that directory.
        # This is necessary because if the sessions_directory_name folder already exists
        # and contains sessions, we don't want to create another sessions_directory_name folder
        # within it, as then the existing sessions won't be available for loading.

        if os.path.basename(user_selected_directory) == self.sessions_directory_name:
            user_selected_directory = os.path.dirname(user_selected_directory)
        else:
            if not os.path.isdir(user_selected_directory):
                os.mkdir(user_selected_directory)

        sessions_directory = os.path.abspath(
            os.path.join(user_selected_directory, self.sessions_directory_name))

        if not os.path.isdir(sessions_directory):
            os.mkdir(sessions_directory)
        self.settings.setValue('sessions_folder', sessions_directory)
        self.sessions_directory = sessions_directory

    def loadSessionNew(self):
        '''Loads a new session'''
        self.createSessionFolder()
        if hasattr(self, 'SessionSelectionWindow'):
            self.SessionSelectionWindow.close()

    def loadSessionLoad(self):
        '''Loads a session from a zip file'''
        folder_name = self.SessionSelectionWindow.widgets['select_session_field'].currentText()
        process_name = f"Loading Session: {folder_name}"

        # Create progress bar which just has an increasing timer:
        self.process_finished = False
        self.createUnknownProgressWindow(process_name)
        config_worker = Worker(self.loadSessionConfig, folder=folder_name)
        config_worker.signals.finished.connect(lambda: self.finishLoadConfig(process_name))
        self.threadpool.start(config_worker)
        self.SessionSelectionWindow.close()

    def createSessionFolder(self):
        '''
        Creates a session folder with the name:
        <app_name>-<date>-<time>
        '''
        date_time = datetime.now().strftime("%d-%m-%Y-%H-%M-%S")
        session_folder_name = f'{self.app_name}-{date_time}'
        session_folder_path = os.path.join(self.sessions_directory, session_folder_name)
        os.mkdir(session_folder_path)
        self.current_session_folder = os.path.abspath(session_folder_path)

    def loadSessionConfig(self, folder, **kwargs):
        '''
        Unzips a session folder and saves the contents of the .json session file to
        self.config.


        Parameters
        ----------
        folder : str
            The folder to load.
        '''

        selected_text = folder
        date_and_time = selected_text.split(' ')[-1]
        selected_folder = ""

        for _, _, f in os.walk(self.sessions_directory):
            for _file in f:
                if date_and_time + '.zip' in _file:
                    selected_folder = os.path.join(self.sessions_directory, _file)
                    break

        shutil.unpack_archive(selected_folder, selected_folder[:-4])
        loaded_folder = selected_folder[:-4]
        self.current_session_folder = loaded_folder

        json_filename = os.path.join(self.current_session_folder, "session.json")

        with open(json_filename) as tmp:
            self.config = json.load(tmp)

    def finishLoadConfig(self, process_name):
        '''
        Called when the config file has been loaded.
        In child classes, you will want to re-implement this function to load
        the GUI elements.'''
        self.finishProcess(process_name)

    def finishProcess(self, process_name):
        '''
        Called when a process has finished.
        Closes the progress bar and sets the process_finished flag to True.
        '''
        self.process_finished = True
        self.progress_windows[process_name].close()

    # Handling Closing And Saving Sessions -------------------------------------

    def closeEvent(self, event):
        '''
        Overwrites the default closeEvent to prompt the user to save the session before closing.

        This method occurs when we call `self.close()`
        or when the user clicks the `X` button on the window.
        '''

        # Need to check that we are not inside the current session folder, as this will either be
        # deleted or moved to a new location.

        if self.current_session_folder in os.path.abspath(os.getcwd()):
            # Move out into the folder that contains the session folders:
            os.chdir(self.sessions_directory)

        if not self.should_really_close:
            self.createSaveWindow(event)
            event.ignore()

    def createSaveWindow(self, event):
        ''''
        Creates a dialog which asks the user if they would like to save the session before closing.

        If the event is a QCloseEvent, then the whole app will always be closed, but the user
        first has an option to save the session.

        If the event is anything else, then the user has the option to save the session, but
        the app will not be closed.

        Parameters
        ----------
        event:
            The event which triggered the dialog.
        '''
        self.should_really_close = False
        dialog = SaveSessionDialog(parent=self, title='Save Session')
        self.SaveWindow = dialog

        # We have 2 instances of the window.
        if isinstance(event, QCloseEvent):
            # This is the case where we are quitting the app and the window asks if we
            # would like to save
            dialog.Cancel.clicked.connect(lambda: self.saveQuitDialogRejected(dialog))
            dialog.Ok.clicked.connect(lambda: self.saveQuitDialogAccepted(dialog))
            dialog.Cancel.setText('Quit without saving')
        else:
            # This is the case where we are just choosing to 'Save' in the file menu
            # so we never quit the app.
            dialog.Cancel.clicked.connect(lambda: self.saveDialogRejected(dialog))
            dialog.Ok.clicked.connect(lambda: self.saveDialogAccepted(dialog))
            dialog.Cancel.setText('Cancel')

        dialog.open()

    def saveDialogAccepted(self, dialog):
        '''
        Called when the user clicks 'Save' in the save dialog.
        This saves the session and then closes the dialog.
        '''
        compress = dialog.widgets['compress_field'].isChecked()
        dialog.close()
        session_name = dialog.widgets['session_name_field'].text()
        self.runSaveSessionWorker(session_name, compress, None)

    def saveDialogRejected(self, dialog):
        '''
        Called when the user clicks 'Cancel' in the save dialog.
        This closes the dialog.
        '''
        dialog.close()

    def saveQuitDialogAccepted(self, dialog):
        '''
        Called when the user clicks 'Save' in the save and quit dialog.
        This saves the session and then closes the app.
        '''
        self.should_really_close = True
        compress = dialog.widgets['compress_field'].isChecked()
        session_name = dialog.widgets['session_name_field'].text()
        dialog.close()
        self.runSaveSessionWorker(session_name, compress, QCloseEvent())

    def saveQuitDialogRejected(self, dialog):
        '''
        Called when the user clicks 'Quit without saving' in the save and quit dialog.
        This closes the app.
        '''
        self.removeTemp() # remove tempdir for this session.
        dialog.close()
        self.should_really_close = True
        self.close()

    def runSaveSessionWorker(self, session_name, compress, event):
        '''
        Runs self.saveSession in a thread.

        When the thread is finished, it calls self.closeSaveWindow, or self.removeTempAndClose
        depending on the type of event.

        If the event is a QCloseEvent, then the whole app will always be closed, by calling
        self.removeTempAndClose.

        If the event is anything else, then the app will not be closed, by calling
        self.closeSaveWindow.
        '''
        process_name = 'Save Session'

        self.process_finished = False
        self.createUnknownProgressWindow(process_name, "Saving", "Saving Session")

        saveSession_worker = Worker(self.saveSession, session_name, compress)
        if isinstance(event, QCloseEvent):
            saveSession_worker.signals.finished.connect(
                lambda: self.removeTempAndClose(process_name))
        else:
            saveSession_worker.signals.finished.connect(lambda: self.closeSaveWindow(process_name))
        self.threadpool.start(saveSession_worker)

    def saveSession(self, session_name, compress, **kwargs):
        '''
        Save the session to a zip file
        The zip file contains the session.json file and the nxs files, and any files
        already in the session folder
        '''
        self.moveSessionFolder(session_name)
        self.saveSessionConfigToJson()
        zip_directory(self.current_session_folder, compress)

    def getSessionConfig(self):
        '''
        Returns the session config, which is a dictionary containing the session name,
        the datetime, the list of files in the session folder, the config of all the
        widgets in the app.

        This must be re-implemented in the child class, to perform the above.
        '''
        return {}

    def moveSessionFolder(self, session_name):
        '''
        Creates a new session folder, and moves the current session folder to it.
        Saves new session folder as self.current_session_folder
        Moves into the new session folder
        '''

        now_string = datetime.now().strftime("%d-%m-%Y-%H-%M")
        new_folder_to_save_to = os.path.join(self.sessions_directory,
                                             session_name + "_" + now_string)
        shutil.move(self.current_session_folder, new_folder_to_save_to)
        self.current_session_folder = new_folder_to_save_to

    def saveSessionConfigToJson(self):
        '''
        Saves the session config to a json file in the session folder: self.current_session_folder
        '''
        session_file = os.path.join(self.current_session_folder, "session.json")

        self.config = self.getSessionConfig()

        self.config['datetime'] = datetime.now().strftime("%d-%m-%Y-%H-%M")

        # Session file will always have the same name.
        # If a session has been reloaded & saved again,
        # this will overwrite the old session file.
        with open(session_file, "w+") as f:
            json.dump(self.config, f)

    def removeTempAndClose(self, process_name):
        '''
        Removes the temp directory for this session, and closes the app.
        '''
        self.removeTemp()
        self.finishProcess(process_name)
        self.should_really_close = True
        self.close()

    def removeTemp(self):
        '''
        Removes the temp directory for this session.
        '''
        if hasattr(self, 'current_session_folder'):
            try:
                shutil.rmtree(self.current_session_folder)
            except PermissionError as e:
                dialog = ErrorDialog(self, message=str(e))
                dialog.exec_()
                self.removeTemp()

    def closeSaveWindow(self, process_name):
        '''
        Closes the save window and moves back to the session folder.
        '''
        # move back to session dir in case we moved
        # out of it for zipping etc. :
        self.finishProcess(process_name)
        self.SaveWindow.close()
